import os
import socket
from multiprocessing import Process
from threading import Thread

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from ftio.parse.metrics import Metrics
from ftio.parse.scales import Scales
from ftio.plot.dash_files.dash_app import IOAnalysisApp
from ftio.plot.print_html import print_html
from ftio.plot.helper import *  
from ftio.plot.plot_error import plot_error_bar, plot_time_error_bar
from ftio.plot.units import find_unit
from ftio.freq.helper import MyConsole


CONSOLE = MyConsole()

def _find_free_port():
    sock = socket.socket()
    sock.bind(("", 0))
    return sock.getsockname()[1]


class plot_core:
    def __init__(self, args):
        self.data = Scales(args)
        self.data.get_data()
        self.names = self.data.names
        self.barprecision = "%{y:.1f}"
        # ? IEEE
        self.width = 900
        self.height = 400

        self.barprecision = "%{y:.0f}"

    def plot_io(self):
        if "dash" in self.data.args.engine.lower():
            self.plot_dash()
        else:
            self.plot_plotly()

    def plot_dash(self):
        """Starts a dash server to display the figures dynamically."""
        app = IOAnalysisApp(self)
        app.run(port=_find_free_port(), debug=False)

    def plot_plotly(self):
        self.f_aw = []
        self.f_ar = []
        self.f_sw = []
        self.f_sr = []
        self.f_t = []
        print("\n\033[1;34mGenerating plots\033[1;0m")
        args = self.data.args
        path = self.data.paths

        if args.render:
            if "stat" in args.render:  # only generate a single image
                args.threaded = False
                if not args.mode:
                    raise Exception("To use render=static, --mode=... must be also set")
        else:
            args.render = "dynamic"

        print("\033[1;32mMultithreaded is on \033[1;0m") if args.threaded else print(
            "\033[1;31mMultithreaded is off \033[1;0m"
        )

        out = print_html(args, path, self.names)
        out.generate_html_start()
        if args.threaded:
            t = []
            t.append(
                Thread(
                    target=self.plot_and_generate_html,
                    args=(
                        "time",
                        out,
                    ),
                )
            )
            t.append(
                Thread(
                    target=self.plot_and_generate_html,
                    args=(
                        "async write",
                        out,
                    ),
                )
            )  # needs comma to be treated as a tuple
            t.append(
                Thread(
                    target=self.plot_and_generate_html,
                    args=(
                        "async read",
                        out,
                    ),
                )
            )
            t.append(
                Thread(
                    target=self.plot_and_generate_html,
                    args=(
                        "sync write",
                        out,
                    ),
                )
            )
            t.append(
                Thread(
                    target=self.plot_and_generate_html,
                    args=(
                        "sync read",
                        out,
                    ),
                )
            )

            for thread in t:
                thread.start()

            for thread in t:
                thread.join()
        else:
            if "stat" in args.render:
                if "async" in args.mode:
                    if "write" in args.mode:
                        self.plot_io_mode("async write", self.data.df_wat, self.data.df_wab)
                    else:
                        self.plot_io_mode("async read", self.data.df_rat, self.data.df_rab)
                elif "sync" in args.mode:
                    if "write" in args.mode:
                        self.plot_io_mode("sync write", self.data.df_wst)
                    else:
                        self.plot_io_mode("sync read", self.data.df_rst)
                else:
                    pass
                exit(0)
            else:
                self.plot_time()
                out.generate_html_core("time.html", self.get_figure("time.html"))
                self.plot_io_mode("async write", self.data.df_wat, self.data.df_wab)
                out.generate_html_core(
                    "write_async.html", self.get_figure("write_async.html")
                )
                self.plot_io_mode("async read", self.data.df_rat, self.data.df_rab)
                out.generate_html_core(
                    "read_async.html", self.get_figure("read_async.html")
                )
                self.plot_io_mode("sync write", self.data.df_wst)
                out.generate_html_core(
                    "write_sync.html", self.get_figure("write_sync.html")
                )
                self.plot_io_mode("sync read", self.data.df_rst)
                out.generate_html_core(
                    "read_sync.html", self.get_figure("read_sync.html")
                )

        out.generate_html_end()
        print("\033[1;32m------------------- done -------------------\n\033[1;0m")

        return [*self.f_aw, *self.f_ar, *self.f_sr, *self.f_sr, *self.f_t]

    # **********************************************************************
    # *                       2. plot Multithreaded
    # **********************************************************************
    def plot_and_generate_html(self, mode, out):
        if "async" in mode:
            if "write" in mode:
                self.plot_io_mode("async write", self.data.df_wat, self.data.df_wab)
                out.generate_html_core(
                    "write_async.html", self.get_figure("write_async.html")
                )
            else:
                self.plot_io_mode("async read", self.data.df_rat, self.data.df_rab)
                out.generate_html_core(
                    "read_async.html", self.get_figure("read_async.html")
                )
        elif "sync" in mode:
            if "write" in mode:
                self.plot_io_mode("sync write", self.data.df_wst)
                out.generate_html_core(
                    "write_sync.html", self.get_figure("write_sync.html")
                )
            else:
                self.plot_io_mode("sync read", self.data.df_rst)
                out.generate_html_core(
                    "read_sync.html", self.get_figure("read_sync.html")
                )
        else:
            self.plot_time()
            out.generate_html_core("time.html", self.get_figure("time.html"))

    # **********************************************************************
    # *                       2. plot
    # **********************************************************************
    def plot_io_mode(self, mode, df_t, df_b=[]):
        """Plots I/O

        Args:
            mode (string): _description_
            df_t (list): list containing throughput
            df_b (list, optional): list containing bandwidth (async). Defaults to [].
        """
        args = self.data.args
        print("\033[1;32m   '-> Creating plot %s\033[1;0m" % (mode))
        f = []
        if isinstance(df_t[1], pd.DataFrame) and df_t[1].empty:
            # Empty throughput by B might exist
            df_t = []
            # empty dataset
            if not df_b or isinstance(df_b[1], pd.DataFrame) and df_b[1].empty:
                self.set_figure(f, mode)
                return

        f.append(go.Figure())
        if df_t:
            f[-1].add_trace(
                go.Box(
                    x=df_t[2]["number_of_ranks"],
                    y=df_t[2]["b_rank_avr"],
                    name="$T_{i,j}$",
                )
            )
        if df_b:
            f[-1].add_trace(
                go.Box(
                    x=df_b[2]["number_of_ranks"],
                    y=df_b[2]["b_rank_sum"],
                    name="$B_{i,j}$",
                )
            )

        if df_t:
            # f.append(go.Figure())
            f[-1].add_trace(
                go.Scatter(
                    x=df_t[0]["number_of_ranks"],
                    y=df_t[0]["harmonic_mean"],
                    mode="lines+markers",
                    name="$hmean(T_{i,j})$",
                    marker_symbol=0,
                    line_dash="dash",
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_t[0]["number_of_ranks"],
                    y=df_t[0]["arithmetic_mean"],
                    mode="lines+markers",
                    name="$amean(T_{i,j})$",
                    marker_symbol=1,
                    line_dash="dash",
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_t[0]["number_of_ranks"],
                    y=df_t[0]["median"],
                    mode="lines+markers",
                    name="$median(T_{i,j})$",
                    marker_symbol=2,
                    line_dash="dash",
                )
            )
        if df_b:
            f[-1].add_trace(
                go.Scatter(
                    x=df_b[0]["number_of_ranks"],
                    y=df_b[0]["harmonic_mean"],
                    mode="lines+markers",
                    name="$hmean(B_{i,j})$",
                    marker_symbol=100,
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_b[0]["number_of_ranks"],
                    y=df_b[0]["arithmetic_mean"],
                    mode="lines+markers",
                    name="$amean(B_{i,j})$",
                    marker_symbol=101,
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_b[0]["number_of_ranks"],
                    y=df_b[0]["median"],
                    mode="lines+markers",
                    name="$median(B_{i,j})$",
                    marker_symbol=102,
                )
            )
        f[-1].update_layout(
            xaxis_title="Ranks",
            yaxis_title="Transfer Rate (B/s)",
            width=self.width,
            height=self.height,
        )
        f[-1] = format_plot(f[-1])

        f.append(go.Figure())
        if df_t:
            f[-1].add_trace(
                go.Box(
                    x=df_t[2]["number_of_ranks"],
                    y=df_t[2]["b_rank_avr"],
                    name="$T_{i,j}$",
                    boxpoints="all",
                    jitter=0.3,
                )
            )
        if df_b:
            f[-1].add_trace(
                go.Box(
                    x=df_b[2]["number_of_ranks"],
                    y=df_b[2]["b_rank_sum"],
                    name="$B_{i,j}$",
                    boxpoints="all",
                    jitter=0.3,
                )
            )
        f[-1].update_layout(
            xaxis_title="Ranks",
            yaxis_title="Transfer Rate (B/s)",
            width=self.width,
            height=self.height,
        )  # ,legend_title_text='Metrics' )
        f[-1] = format_plot(f[-1])

        f.append(go.Figure())
        if df_t:
            f[-1].add_trace(
                go.Scatter(
                    x=df_t[0]["number_of_ranks"],
                    y=df_t[0]["harmonic_mean"],
                    mode="lines+markers",
                    name="$hmean(T_{i,j})$",
                    marker_symbol=0,
                    line_dash="dash",
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_t[0]["number_of_ranks"],
                    y=df_t[0]["arithmetic_mean"],
                    mode="lines+markers",
                    name="$amean(T_{i,j})$",
                    marker_symbol=1,
                    line_dash="dash",
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_t[0]["number_of_ranks"],
                    y=df_t[0]["median"],
                    mode="lines+markers",
                    name="$median(T_{i,j})$",
                    marker_symbol=2,
                    line_dash="dash",
                )
            )
        if df_b:
            f[-1].add_trace(
                go.Scatter(
                    x=df_b[0]["number_of_ranks"],
                    y=df_b[0]["harmonic_mean"],
                    mode="lines+markers",
                    name="$hmean(B_{i,j})$",
                    marker_symbol=100,
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_b[0]["number_of_ranks"],
                    y=df_b[0]["arithmetic_mean"],
                    mode="lines+markers",
                    name="$amean(B_{i,j})$",
                    marker_symbol=101,
                )
            )
            f[-1].add_trace(
                go.Scatter(
                    x=df_b[0]["number_of_ranks"],
                    y=df_b[0]["median"],
                    mode="lines+markers",
                    name="$median(B_{i,j})$",
                    marker_symbol=102,
                )
            )
        f[-1].update_layout(
            xaxis_title="Ranks",
            yaxis_title="Transfer Rate (B/s)",
            width=self.width,
            height=self.height,
        )
        f[-1] = format_plot(f[-1])

        _ = go.Figure(
            layout=dict(template="plotly")
        )  #!fixes plotly error for default values
        if df_t:
            fig_tmp = px.histogram(
                df_t[1],
                x="b_overlap_avr",
                color="number_of_ranks",
                marginal="box",
                labels={
                    "b_overlap_avr": "Throughput  (B/s)",
                    "number_of_ranks": "Ranks",
                },
            )  # "number_of_ranks": "idth (cm)","number_of_ranks": "Rank"
            fig_tmp.update_layout(
                width=self.width,
                height=self.height,
                title="Distribution",
                xaxis_title="Throughput (B/s)",
                yaxis_title="Count",
            )
            fig_tmp = format_plot(fig_tmp)
            f.append(fig_tmp)

        if df_b:
            fig_tmp = px.histogram(
                df_b[1],
                x="b_overlap_sum",
                color="number_of_ranks",
                marginal="box",
                labels={
                    "b_overlap_sum": "Bandwidth  (B/s)",
                    "number_of_ranks": "Ranks",
                },
            )  # "number_of_ranks": "idth (cm)","number_of_ranks": "Rank"
            fig_tmp.update_layout(
                width=self.width,
                height=self.height,
                title="Distribution",
                xaxis_title="Bandwidth (B/s)",
                yaxis_title="Count",
            )
            fig_tmp = format_plot(fig_tmp)
            f.append(fig_tmp)

        io_stats = Metrics(args)
        if df_t:
            ranks = df_t[0]["number_of_ranks"].sort_index()
        elif df_b:
            ranks = df_b[0]["number_of_ranks"].sort_index()
        for i in pd.unique(ranks):
            try:
                i = int(i)
                index = df_t[1]["number_of_ranks"].isin([i])
                index_ind = df_t[3]["number_of_ranks"].isin([i])
            except:
                index = df_t[1]["number_of_ranks"].astype(str).isin([i])
                index_ind = df_t[3]["number_of_ranks"].astype(str).isin([i])
            if (
                len(df_t[1]["file_index"][index]) != 0
                or len(df_t[3]["file_index"][index_ind]) != 0
            ):
                for j in range(
                    int(df_t[1]["file_index"][index].min()),
                    int(df_t[1]["file_index"][index].max() + 1),
                ):
                    f.append(go.Figure())
                    index2 = df_t[1]["file_index"][index].isin([j])
                    index2_ind = df_t[3]["file_index"][index_ind].isin([j])
                    # finder the order and unit of the plots on y-axis
                    unit, order = find_unit(df_t,index,index2,index_ind,index2_ind,args)
                    # # compatibility mode with earlier version
                    # CONSOLE.info(f"[yellow]WARNING:[/] Compatibility Mode")
                    # unit, order = find_unit(df_t,index,index2,index_ind,index2_ind,args,pow(10,6)) 
                    # order = order*pow(10,6)
                    # ? Avr plot
                    if args.avr:
                        if df_t:
                            f[-1].add_trace(
                                go.Scatter(
                                    x=df_t[1]["t_overlap"][index][index2],
                                    y=df_t[1]["b_overlap_avr"][index][index2]*order,
                                    mode="lines",
                                    name="$T_A$", #name="$T$", # For BW limit paper
                                    line={"shape": "hv"},
                                    fill="tozeroy",
                                )
                            )

                        if df_b:
                            f[-1].add_trace(
                                go.Scatter(
                                    x=df_b[1]["t_overlap"][index][index2],
                                    y=df_b[1]["b_overlap_avr"][index][index2]*order,
                                    mode="lines",
                                    name="$B_A$", #name="$B_L$", # For BW limit paper
                                    line={"shape": "hv"},
                                    fill="tozeroy",
                                )
                            )

                    # ? Sum plot
                    if args.sum:
                        if df_t:
                            f[-1].add_trace(
                                go.Scatter(
                                    x=df_t[1]["t_overlap"][index][index2],
                                    y=df_t[1]["b_overlap_sum"][index][index2]*order,
                                    mode="lines",
                                    name="$T_S$",
                                    line={"shape": "hv"},
                                    fill="tozeroy",
                                    )
                                )  

                        if df_b:
                            f[-1].add_trace(
                                go.Scatter(
                                    x=df_b[1]["t_overlap"][index][index2],
                                    y=df_b[1]["b_overlap_sum"][index][index2]*order,
                                    mode="lines",
                                    name="$B_S$",#name="$B$", # For BW limit paper
                                    line={"shape": "hv"},
                                    fill="tozeroy",
                                )
                            )

                    # ? ind plot
                    if args.ind:
                        if df_t:
                            f[-1].add_trace(
                                go.Scatter(
                                    x=df_t[3]["t_overlap_ind"][index_ind][index2_ind],
                                    y=df_t[3]["b_overlap_ind"][index_ind][index2_ind]*order,
                                    mode="lines",
                                    name="$T_E$",
                                    line={"shape": "hv"},
                                    fill="tozeroy",
                                )
                            ) 

                        if df_b:
                            f[-1].add_trace(
                                go.Scatter(
                                    x=df_b[3]["t_overlap_ind"][index_ind][index2_ind],
                                    y=df_b[3]["b_overlap_ind"][index_ind][index2_ind]*order,
                                    mode="lines",
                                    name="$B_E$",
                                    line={"shape": "hv"},
                                    fill="tozeroy",
                                )
                            )  

                    f[-1].update_layout(
                        barmode="stack",
                        xaxis_title="Time (s)",
                        yaxis_title=f"Transfer Rate ({unit})",
                        width=self.width,
                        height=self.height,
                        showlegend=True,
                        title=f"{i} Ranks (Run {j})",
                    )
                    f[-1] = format_plot(f[-1])
                    
                    if self.names and (args.avr or args.sum or args.ind):
                        f[-1].update_layout(
                            title=f"{i} Ranks (Run {j}: {self.names[j]})",
                        )

                    if args.avr or args.sum or args.ind:
                        if args.zoom:
                            f[-1].update_layout(yaxis_range=[0, args.zoom])
                        else:
                            f[-1].update_layout(yaxis_rangemode="nonnegative")
                        f[-1].update_xaxes(
                            ticks="outside",
                            tickcolor="black",
                            ticklen=6,
                            minor=dict(
                                ticklen=3,
                                tickcolor="black",
                                tickmode="auto",
                                nticks=10,
                                showgrid=True,
                            ),
                        )
                        f[-1].update_yaxes(
                            ticks="outside",
                            tickcolor="black",
                            ticklen=6,
                            minor=dict(
                                ticklen=3,
                                tickcolor="black",
                                tickmode="auto",
                                nticks=10,
                                showgrid=True,
                            ),
                        )

                    #! compute statistics
                    if df_b:
                        io_stats.add(
                            i,
                            j,
                            df_t[1][index][index2],
                            df_t[3][index_ind][index2_ind],
                            df_b[1][index][index2],
                            df_b[3][index_ind][index2_ind],
                        )
                    else:
                        if df_t:
                            io_stats.add(
                                i,
                                j,
                                df_t[1][index][index2],
                                df_t[3][index_ind][index2_ind],
                            )

            if "stat" in args.render:
                path = os.path.join(os.getcwd(), f"exported_images_{i}")
                if not os.path.exists(path):
                    os.mkdir(path)
                if not self.names:
                    self.names.append(str(i))
                self.names = [x.replace(".", "_") for x in self.names]
                procs = []
                for fig in f:
                    procs.append(
                        Process(target=save_fig, args=(fig, f, path, self.names[-1]))
                    )
                    procs[-1].start()

                while procs:
                    for p in procs:
                        if p.is_alive():
                            pass
                        else:
                            p.join()
                            procs.remove(p)

                print(f"Images  saved to {path}\nExciting")
                exit(0)

        #! plot overlap statistics
        if self.nRun >= 1:
            io_stats.get_data()
            labels = ["A", "S", "E"]
            for count, type in enumerate(["average", "sum", "ind"]):
                f.append(go.Figure())
                f[-1].update_layout(width=self.width, height=self.height)
                f[-1].update_xaxes(
                    minor=dict(ticklen=6, tickcolor="black", showgrid=True)
                )
                for i in ["max", "min", "median", "hmean", "amean"]:
                    if df_t:
                        f[-1].add_trace(
                            go.Scatter(
                                x=io_stats.get("t_%s" % type[0:3], "number_of_ranks"),
                                y=io_stats.get("t_%s" % type[0:3], i),
                                name="$\\text{%s}(T_{%s})$" % (i, labels[count]),
                                hovertemplate="<b>Throughput</b><br>Ranks: %{x:i}<br>"
                                + i
                                + ": %{y:.2f}<br>",
                            )
                        )
                    if df_b:
                        f[-1].add_trace(
                            go.Scatter(
                                x=io_stats.get("b_%s" % type[0:3], "number_of_ranks"),
                                y=io_stats.get("b_%s" % type[0:3], i),
                                name="$\\text{%s}(B_{%s})$" % (i, labels[count]),
                                hovertemplate="<b>Bandwidth</b><br>Ranks: %{x:i}<br>"
                                + i
                                + ": %{y:.2f}<br>",
                            )
                        )
                f[-1].update_layout(
                    barmode="stack",
                    xaxis_title="Ranks",
                    yaxis_title="Transfer Rate (B/s)",
                    title="Overlap Statistics: %s" % type.capitalize(),
                )
                f[-1] = format_plot(f[-1])

            f_1 = plot_error_bar(io_stats.t_avr, "T_{A}")
            f_2 = plot_error_bar(io_stats.t_sum, "T_{S}")
            f_3 = plot_error_bar(io_stats.t_ind, "T_{E}")
            if df_b:
                f_1 = plot_error_bar(io_stats.b_avr, "B_{A}", f_1)
                f_2 = plot_error_bar(io_stats.b_sum, "B_{S}", f_2)
                f_3 = plot_error_bar(io_stats.b_ind, "B_{E}", f_3)
            f.extend([f_1, f_2, f_3])

        self.set_figure(f, mode)

    # **********************************************************************
    # *                       3. plot_Time
    # **********************************************************************
    def plot_time(self):
        # ? Init
        print("\033[1;32m   '-> Creating plot %s\033[1;0m" % ("I/O time"))
        colors = px.colors.qualitative.Plotly
        self.nRun = len(pd.unique(self.data.df_time["file_index"]))
        colors = px.colors.qualitative.Plotly
        symbols = [
            "square",
            "circle",
            "cross",
            "star-triangle-down",
            "hexagon",
            "x",
            "diamond",
            "star-triangle-up",
        ]
        markeredgecolor = "DarkSlateGrey"
        self.width_increase = 0
        self.height = 1.0 * self.height
        if self.nRun == 1:
            x = self.data.df_time["number_of_ranks"].astype(str)

        else:
            self.data.df_time = self.data.df_time.sort_values(
                by=["number_of_ranks", "file_index"]
            )
            x = [
                self.data.df_time["number_of_ranks"].astype(str),
                self.data.df_time["file_index"].astype(str),
            ]
            # x = [self.data.df_time.sort_values(by=["number_of_ranks","file_index"])["file_index"].astype(str), self.data.df_time.sort_values(by=["number_of_ranks","file_index"])["file_index"].astype(str)]

            self.width = self.width
            self.width_increase = (self.nRun - 1) * self.width / 5
            # self.width_increase = -self.width + self.nRun * len(self.data.df_time['number_of_ranks'])*self.width/24

        # ? Generate Figures
        self.f_t.append(add_fig_col(self.nRun))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_total"][index],
                    mode="lines+markers",
                    fill="tozeroy",
                    name="Total",
                    legendgroup="Total",
                    marker=dict(
                        symbol=symbols[0], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[0],
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_agg"][index],
                    mode="lines+markers",
                    fill="tozeroy",
                    name="App",
                    legendgroup="App",
                    marker=dict(
                        symbol=symbols[1], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[1],
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_overhead"][index],
                    mode="lines+markers",
                    fill="tozeroy",
                    name="Overhead",
                    legendgroup="Overhead",
                    marker=dict(
                        symbol=symbols[2], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[2],
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].update_xaxes(title_text="Ranks", row=1, col=i + 1)
        self.f_t[-1].update_layout(
            xaxis_title="Ranks",
            yaxis_title="Time (s)",
            width=self.width + self.width_increase,
            height=self.height,
            title="Time Distribution",
        )
        self.f_t[-1].update_layout(hovermode="x", legend_tracegroupgap=1)
        self.f_t[-1].update_xaxes(showspikes=True, spikethickness=1, spikecolor="black")
        self.f_t[-1].update_yaxes(showspikes=True, spikethickness=1, spikecolor="black")
        self.f_t[-1] = format_plot(self.f_t[-1])

        modes = ["delta_t_total", "delta_t_agg", "delta_t_overhead"]
        names = ["Total", "App", "Overhead"]
        self.f_t.append(
            plot_time_error_bar(
                self.data.df_time, modes, names, colors, symbols, markeredgecolor
            )
        )

        self.f_t.append(go.Figure())
        if self.nRun != 1:
            self.f_t[-1].update_layout(barmode="relative")
            self.f_t[-1].update_layout(
                legend=dict(
                    orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                )
            )
        self.f_t[-1].add_trace(
            go.Bar(
                x=x,
                y=self.data.df_time["delta_t_agg"]
                / self.data.df_time["delta_t_total"]
                * 100,
                text=self.data.df_time["delta_t_agg"]
                / self.data.df_time["delta_t_total"]
                * 100,
                name="App",
                textposition="inside",
                textangle=0,
                texttemplate=self.barprecision,
                textfont=dict(color="white"),
            )
        )
        self.f_t[-1].add_trace(
            go.Bar(
                x=x,
                y=self.data.df_time["delta_t_overhead"]
                / self.data.df_time["delta_t_total"]
                * 100,
                text=self.data.df_time["delta_t_overhead"]
                / self.data.df_time["delta_t_total"]
                * 100,
                name="Overhead",
                textposition="inside",
                textangle=0,
                texttemplate=self.barprecision,
                textfont=dict(color="white"),
            )
        )
        self.f_t[-1].update_layout(
            xaxis_title_text="Ranks",
            yaxis_title_text="Percentage (%)",
            height=self.height,
            title="Total time",
            barmode="stack",
        )  # ,uniformtext_minsize=12, uniformtext_mode='show')
        self.f_t[-1].update_layout(barmode="relative")
        self.f_t[-1].update_yaxes(range=[0, 100])
        self.f_t[-1] = format_plot(self.f_t[-1])

        self.f_t.append(go.Figure())
        delmiter = "<br>"
        if self.nRun != 1:
            delmiter = " "
            self.f_t[-1].update_layout(barmode="relative")
            self.f_t[-1].update_layout(
                legend=dict(
                    orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                )
            )
        self.f_t[-1].add_trace(
            go.Bar(
                x=x,
                y=self.data.df_time["delta_t_com"]
                / self.data.df_time["delta_t_total"]
                * 100,
                text=self.data.df_time["delta_t_com"]
                / self.data.df_time["delta_t_total"]
                * 100,
                name="Compute",
                textposition="inside",
                textangle=0,
                marker_color=colors[0],
                texttemplate=self.barprecision,
                textfont=dict(color="white"),
            )
        )
        self.f_t[-1].add_trace(
            go.Bar(
                x=x,
                y=self.data.df_time["delta_t_agg_io"]
                / self.data.df_time["delta_t_total"]
                * 100,
                text=self.data.df_time["delta_t_agg_io"]
                / self.data.df_time["delta_t_total"]
                * 100,
                name="Visible I/O",
                textposition="inside",
                textangle=0,
                marker_color=colors[1],
                texttemplate=self.barprecision,
                textfont=dict(color="white"),
            )
        )
        if (
            "delta_t_overhead_post_runtime" in self.data.df_time.head()
            and "delta_t_overhead_peri_runtime" in self.data.df_time.head()
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_overhead_peri_runtime"]
                    / self.data.df_time["delta_t_total"]
                    * 100,
                    text=self.data.df_time["delta_t_overhead_peri_runtime"]
                    / self.data.df_time["delta_t_total"]
                    * 100,
                    name="Overhead%speri-run" % delmiter,
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[7],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_overhead_post_runtime"]
                    / self.data.df_time["delta_t_total"]
                    * 100,
                    text=self.data.df_time["delta_t_overhead_post_runtime"]
                    / self.data.df_time["delta_t_total"]
                    * 100,
                    name="Overhead%spost-run" % delmiter,
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[8],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        else:
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_overhead"]
                    / self.data.df_time["delta_t_total"]
                    * 100,
                    text=self.data.df_time["delta_t_overhead"]
                    / self.data.df_time["delta_t_total"]
                    * 100,
                    name="Overhead",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[9],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        self.f_t[-1].update_layout(
            xaxis_title_text="Ranks",
            yaxis_title_text="Percentage (%)",
            width=self.width + self.width_increase,
            height=self.height,
            title="Total Time",
            barmode="stack",
        )  # ,uniformtext_minsize=12, uniformtext_mode='show')
        self.f_t[-1].update_yaxes(range=[0, 100])
        self.f_t[-1] = format_plot(self.f_t[-1])

        self.f_t.append(add_fig_col(self.nRun))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_agg"][index],
                    mode="lines+markers",
                    fill="tozeroy",
                    name="app",
                    marker=dict(
                        symbol=symbols[1], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[0],
                    legendgroup="app",
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_com"][index],
                    mode="lines+markers",
                    fill="tozeroy",
                    name="Compute",
                    marker=dict(
                        symbol=symbols[3], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[1],
                    legendgroup="Compute",
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_agg_io"][index],
                    mode="lines+markers",
                    fill="tozeroy",
                    name="Visible I/O",
                    marker=dict(
                        symbol=symbols[4], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[2],
                    legendgroup="I/O",
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].update_xaxes(title_text="Ranks", row=1, col=i + 1)
        self.f_t[-1].update_layout(
            legend_tracegroupgap=1,
            xaxis_title="Ranks",
            yaxis_title="Time (s)",
            width=self.width + self.width_increase,
            height=self.height,
            title="Application Time",
        )
        self.f_t[-1].update_layout(hovermode="x", legend_tracegroupgap=1)
        self.f_t[-1] = format_plot(self.f_t[-1])

        modes = ["delta_t_agg", "delta_t_com", "delta_t_agg_io"]
        names = ["App", "Compute", "I/O"]
        self.f_t.append(
            plot_time_error_bar(
                self.data.df_time, modes, names, colors, symbols, markeredgecolor
            )
        )
        self.f_t[-1] = format_plot(self.f_t[-1])
        self.f_t[-1].update_yaxes(type="log", row=1, col=1, showspikes=True)

        self.f_t.append(go.Figure())
        if self.nRun != 1:
            self.f_t[-1].update_layout(barmode="relative")
            self.f_t[-1].update_layout(
                legend=dict(
                    orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                )
            )
        self.f_t[-1].add_trace(
            go.Bar(
                x=x,
                y=self.data.df_time["delta_t_com"]
                / self.data.df_time["delta_t_agg"]
                * 100,
                text=self.data.df_time["delta_t_com"]
                / self.data.df_time["delta_t_agg"]
                * 100,
                name="Compute",
                textposition="inside",
                textangle=0,
                texttemplate=self.barprecision,
                textfont=dict(color="white"),
            )
        )
        self.f_t[-1].add_trace(
            go.Bar(
                x=x,
                y=self.data.df_time["delta_t_agg_io"]
                / self.data.df_time["delta_t_agg"]
                * 100,
                text=self.data.df_time["delta_t_agg_io"]
                / self.data.df_time["delta_t_agg"]
                * 100,
                name="I/O",
                textposition="inside",
                textangle=0,
                texttemplate=self.barprecision,
                textfont=dict(color="white"),
            )
        )

        self.f_t[-1].update_layout(
            xaxis_title_text="Ranks",
            yaxis_title_text="Percentage (%)",
            width=self.width + self.width_increase,
            height=self.height,
            title="Application Time",
            barmode="stack",
        )  # ,uniformtext_minsize=12, uniformtext_mode='show')
        self.f_t[-1].update_yaxes(range=[0, 100])
        self.f_t[-1] = format_plot(self.f_t[-1])

        # ? I/O to compute ration
        self.f_t.append(add_fig_col(self.nRun))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["delta_t_com"][index],
                    y=self.data.df_time["delta_t_agg_io"][index],
                    mode="lines+markers",
                    name="Compute to I/O time",
                    legendgroup="Compute",
                    marker=dict(
                        symbol=symbols[3], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[0],
                    showlegend=legend_fix(self, i),
                    text=self.data.df_time["number_of_ranks"][index],
                    hovertemplate="<b>Ranks: %{text}</b><br>"
                    + "Comp. time : %{x:.2f}<br>"
                    + "I/O time: %{y:.2}"
                    + "<extra></extra>",
                ),
                row=1,
                col=i + 1,
            )  # , secondary_y=False)
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["delta_t_com"][index],
                    y=self.data.df_time["delta_t_com"][index],
                    mode="lines+markers",
                    name="Ref",
                    legendgroup="Ref",
                    marker=dict(
                        symbol=symbols[4], line=dict(width=1, color=markeredgecolor)
                    ),
                    line_dash="dash",
                    marker_color=colors[1],
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )  # , secondary_y=False)
            # self.f_t[-1].add_trace(go.Scatter(x=self.data.df_time['delta_t_com'][index],     y=self.data.df_time['delta_t_agg_io'][index], mode = 'lines+markers', name = 'Compute to I/O time', marker_symbol=0, marker_color=colors[0], showlegend = False), row=1, col=i+1), secondary_y=True)
            self.f_t[-1].update_xaxes(title_text="Compute time (s)", row=1, col=i + 1)
        self.f_t[-1].update_layout(
            legend_tracegroupgap=1,
            xaxis_title="Compute time (s)",
            yaxis_title="I/O time (s)",
            width=self.width + self.width_increase,
            height=self.height,
            title="Compute to I/O time",
        )
        self.f_t[-1].update_yaxes(
            title_text="Ranks",
            secondary_y=True,
            ticktext=self.data.df_time["number_of_ranks"][index],
            tickvals=self.data.df_time["delta_t_agg_io"][index],
            showspikes=True,
        )
        self.f_t[-1] = format_plot(self.f_t[-1])

        #! I/O instensity for the malleability manager
        self.f_t.append(add_fig_col(self.nRun))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_agg_io"][index]
                    / self.data.df_time["delta_t_com"][index],
                    mode="lines+markers",
                    name="I/O time to compute",
                    legendgroup="IO1",
                    marker=dict(
                        symbol=symbols[3], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[0],
                    showlegend=legend_fix(self, i),
                    text=self.data.df_time["number_of_ranks"][index],
                    hovertemplate="<b>Ranks: %{text}</b><br>"
                    + "Ratio: %{y:.2}"
                    + "<extra></extra>",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_agg_io"][index]
                    / self.data.df_time["delta_t_agg"][index],
                    mode="lines+markers",
                    name="I/O time to total",
                    legendgroup="IO2",
                    marker=dict(
                        symbol=symbols[5], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[2],
                    showlegend=legend_fix(self, i),
                    text=self.data.df_time["number_of_ranks"][index],
                    hovertemplate="<b>Ranks: %{text}</b><br>"
                    + "Ratio: %{y:.2}"
                    + "<extra></extra>",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=np.repeat(0.5, len(self.data.df_time["number_of_ranks"][index])),
                    mode="lines+markers",
                    name="Ref",
                    legendgroup="Ref",
                    marker=dict(
                        symbol=symbols[4], line=dict(width=1, color=markeredgecolor)
                    ),
                    line_dash="dash",
                    marker_color=colors[1],
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )  # , secondary_y=False)
            # self.f_t[-1].add_trace(go.Scatter(x=self.data.df_time['delta_t_com'][index],     y=self.data.df_time['delta_t_agg_io'][index], mode = 'lines+markers', name = 'Compute to I/O time', marker_symbol=0, marker_color=colors[0], showlegend = False), row=1, col=i+1), secondary_y=True)
            self.f_t[-1].update_xaxes(title_text="Ranks", row=1, col=i + 1)
        self.f_t[-1].update_layout(
            legend_tracegroupgap=1,
            xaxis_title="Ranks",
            yaxis_title="I/O intensity",
            width=self.width + self.width_increase,
            height=self.height,
            title="I/O intensity",
        )
        self.f_t[-1] = format_plot(self.f_t[-1])
        self.f_t[-1].update_yaxes(
            title_text="Ranks",
            secondary_y=True,
            ticktext=self.data.df_time["number_of_ranks"][index],
            tickvals=self.data.df_time["delta_t_agg_io"][index],
            showspikes=True,
        )
        self.f_t[-1].update_xaxes(showspikes=True)
        self.f_t[-1].update_yaxes(range=[0, 1])

        # ? I/O instensity error bar
        modes = ["delta_t_agg_io / delta_t_com", "delta_t_agg_io / delta_t_agg"]
        names = ["I/O to compute", "I/O to total"]
        self.f_t.append(
            plot_time_error_bar(
                self.data.df_time, modes, names, colors, symbols, markeredgecolor
            )
        )

        self.f_t.append(go.Figure())
        colors = px.colors.qualitative.Plotly
        y = (
            self.data.df_time["delta_t_com"]
            - (self.data.df_time["delta_t_ara"] - self.data.df_time["delta_t_ar_lost"])
            - (self.data.df_time["delta_t_awa"] - self.data.df_time["delta_t_aw_lost"])
        )
        if self.nRun != 1:
            self.f_t[-1].update_layout(barmode="relative")
            self.f_t[-1].update_layout(
                legend=dict(
                    orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                )
            )
        if not y.empty and not all(y == 0):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=y / self.data.df_time["delta_t_agg"] * 100,
                    text=y / self.data.df_time["delta_t_agg"] * 100,
                    name="Compute (I/O free)",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[0],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        if not self.data.df_time["delta_t_awa"].empty and not all(
            self.data.df_time["delta_t_awa"] == 0
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=(
                        self.data.df_time["delta_t_awa"]
                        - self.data.df_time["delta_t_aw_lost"]
                    )
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    text=(
                        self.data.df_time["delta_t_awa"]
                        - self.data.df_time["delta_t_aw_lost"]
                    )
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    name="Async write exploit",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[1],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        if not self.data.df_time["delta_t_ara"].empty and not all(
            self.data.df_time["delta_t_ara"] == 0
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=(
                        self.data.df_time["delta_t_ara"]
                        - self.data.df_time["delta_t_ar_lost"]
                    )
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    text=(
                        self.data.df_time["delta_t_ara"]
                        - self.data.df_time["delta_t_ar_lost"]
                    )
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    name="Async read exploit",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[2],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        if not self.data.df_time["delta_t_awa"].empty and not all(
            self.data.df_time["delta_t_awa"] == 0
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_aw_lost"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    text=self.data.df_time["delta_t_aw_lost"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    name="Async write lost",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[3],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        if not self.data.df_time["delta_t_ara"].empty and not all(
            self.data.df_time["delta_t_ara"] == 0
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_ar_lost"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    text=self.data.df_time["delta_t_ar_lost"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    name="Async read lost",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[4],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        if not self.data.df_time["delta_t_sw"].empty and not all(
            self.data.df_time["delta_t_sw"] == 0
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_sw"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    text=self.data.df_time["delta_t_sw"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    name="Sync write",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[5],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        if not self.data.df_time["delta_t_sr"].empty and not all(
            self.data.df_time["delta_t_sr"] == 0
        ):
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_sr"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    text=self.data.df_time["delta_t_sr"]
                    / self.data.df_time["delta_t_agg"]
                    * 100,
                    name="Sync read",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[6],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
        self.f_t[-1].update_layout(
            xaxis_title="Ranks",
            yaxis_title="Percentage (%)",
            width=self.width + self.width_increase,
            height=1.3 * self.height,
            title="Detailed Application Time",
            barmode="stack",
            margin=dict(t=200),
        )
        self.f_t[-1] = format_plot(self.f_t[-1])

        self.f_t.append(add_fig_col(self.nRun))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_sr"][index],
                    mode="lines+markers",
                    name="Sync read",
                    marker=dict(
                        symbol=symbols[0], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[0],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_sr",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_ara"][index],
                    mode="lines+markers",
                    name="Async read act.",
                    marker=dict(
                        symbol=symbols[1], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[1],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_ara",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_arr"][index],
                    mode="lines+markers",
                    name="Async read req.",
                    marker=dict(
                        symbol=symbols[2], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[2],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_arr",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_ar_lost"][index],
                    mode="lines+markers",
                    name="Async read lost",
                    marker=dict(
                        symbol=symbols[3], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[3],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_ar_lost",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_sw"][index],
                    mode="lines+markers",
                    name="Sync write",
                    marker=dict(
                        symbol=symbols[4], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[4],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_sw",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_awa"][index],
                    mode="lines+markers",
                    name="Async write act.",
                    marker=dict(
                        symbol=symbols[5], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[5],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_awa",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_awr"][index],
                    mode="lines+markers",
                    name="Async write req.",
                    marker=dict(
                        symbol=symbols[6], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[6],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_awr",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_aw_lost"][index],
                    mode="lines+markers",
                    name="Async write lost",
                    marker=dict(
                        symbol=symbols[7], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[7],
                    showlegend=legend_fix(self, i),
                    legendgroup="delta_t_aw_lost",
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].update_xaxes(title_text="Ranks", row=1, col=i + 1)
        self.f_t[-1].update_layout(
            legend_tracegroupgap=1,
            xaxis_title="Ranks",
            yaxis_title="Time (s)",
            width=self.width + self.width_increase,
            height=self.height,
            title="I/O time",
        )
        self.f_t[-1] = format_plot(self.f_t[-1])

        self.f_t.append(add_fig_row(self.nRun, onefig=True))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_sr"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_sr",
                    legendgrouptitle_text="Sync read",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[0], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_ara"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_ara",
                    legendgrouptitle_text="Async read act.",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[1], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_arr"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_arr",
                    legendgrouptitle_text="Async read req.",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[2], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_ar_lost"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_ar_lost",
                    legendgrouptitle_text="Async read lost.",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[3], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_sw"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_sw",
                    legendgrouptitle_text="Sync write ",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[4], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_awa"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_awa",
                    legendgrouptitle_text="Async write act.",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[5], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_awr"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_awr",
                    legendgrouptitle_text="Async write req.",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[6], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_aw_lost"][index],
                    mode="lines+markers",
                    legendgroup="delta_t_aw_lost",
                    legendgrouptitle_text="Async lost",
                    name="run %i" % i,
                    marker=dict(
                        symbol=symbols[7], line=dict(width=1, color=markeredgecolor)
                    ),
                ),
                row=1,
                col=1,
            )
        self.f_t[-1].update_layout(
            legend_tracegroupgap=1,
            xaxis_title="Ranks",
            yaxis_title="Time (s)",
            width=self.width + self.width_increase,
            height=self.height,
            title="I/O time",
        )
        self.f_t[-1] = format_plot(self.f_t[-1])

        #! Error bar
        # modes = ["delta_t_sr", "delta_t_sw",  "delta_t_ar_lost", "delta_t_aw_lost",  "delta_t_ara",     "delta_t_awa",     "delta_t_arr",      "delta_t_awr"]
        # names = ["Sync read",  "Sync write ", "Async read lost", "Async write lost", "Async read act.", "Async write act.", "Async read req.",  "Async write req.", ]
        modes = [
            "delta_t_sr",
            "delta_t_ara",
            "delta_t_arr",
            "delta_t_ar_lost",
            "delta_t_sw",
            "delta_t_awa",
            "delta_t_awr",
            "delta_t_aw_lost",
        ]
        names = [
            "Sync read",
            "Async read act.",
            "Async read req.",
            "Async read lost",
            "Sync write ",
            "Async write act.",
            "Async write req.",
            "Async write lost",
        ]
        self.f_t.append(
            plot_time_error_bar(
                self.data.df_time, modes, names, colors, symbols, markeredgecolor
            )
        )

        self.f_t.append(add_fig_col(self.nRun))
        for i in range(0, self.nRun):
            index = self.data.df_time["file_index"].isin([i])
            self.f_t[-1].add_trace(
                go.Scatter(
                    x=self.data.df_time["number_of_ranks"][index],
                    y=self.data.df_time["delta_t_overhead"][index],
                    mode="lines+markers",
                    name="Overhead",
                    legendgroup="delta_t_overhead",
                    marker=dict(
                        symbol=symbols[0], line=dict(width=1, color=markeredgecolor)
                    ),
                    marker_color=colors[0],
                    showlegend=legend_fix(self, i),
                ),
                row=1,
                col=i + 1,
            )
            self.f_t[-1].update_xaxes(title_text="Ranks", row=1, col=i + 1)
            if "delta_t_overhead_post_runtime" in self.data.df_time.head():
                self.f_t[-1].add_trace(
                    go.Scatter(
                        x=self.data.df_time["number_of_ranks"][index],
                        y=self.data.df_time["delta_t_overhead_post_runtime"][index],
                        mode="lines+markers",
                        name="Post runtime",
                        legendgroup="delta_t_overhead_post_runtime",
                        marker=dict(
                            symbol=symbols[1], line=dict(width=1, color=markeredgecolor)
                        ),
                        marker_color=colors[1],
                        showlegend=legend_fix(self, i),
                    ),
                    row=1,
                    col=i + 1,
                )
            if "delta_t_overhead_peri_runtime" in self.data.df_time.head():
                self.f_t[-1].add_trace(
                    go.Scatter(
                        x=self.data.df_time["number_of_ranks"][index],
                        y=self.data.df_time["delta_t_overhead_peri_runtime"][index],
                        mode="lines+markers",
                        name="During runtime",
                        legendgroup="delta_t_overhead_peri_runtime",
                        marker=dict(
                            symbol=symbols[2], line=dict(width=1, color=markeredgecolor)
                        ),
                        marker_color=colors[2],
                        showlegend=legend_fix(self, i),
                    ),
                    row=1,
                    col=i + 1,
                )
        self.f_t[-1].update_layout(
            legend_tracegroupgap=1,
            xaxis_title="Ranks",
            yaxis_title="Time (s)",
            width=self.width + self.width_increase,
            height=self.height,
            title="Overhead",
        )
        self.f_t[-1] = format_plot(self.f_t[-1])

        if (
            "delta_t_overhead_post_runtime" in self.data.df_time.head()
            and "delta_t_overhead_peri_runtime" in self.data.df_time.head()
        ):
            self.f_t.append(go.Figure())
            if self.nRun != 1:
                self.f_t[-1].update_layout(barmode="relative")
                self.f_t[-1].update_layout(
                    legend=dict(
                        orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                    )
                )
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_overhead_peri_runtime"]
                    / self.data.df_time["delta_t_overhead"]
                    * 100,
                    text=self.data.df_time["delta_t_overhead_peri_runtime"]
                    / self.data.df_time["delta_t_overhead"]
                    * 100,
                    name="During runtime",
                    textposition="inside",
                    textangle=0,
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_overhead_post_runtime"]
                    / self.data.df_time["delta_t_overhead"]
                    * 100,
                    text=self.data.df_time["delta_t_overhead_post_runtime"]
                    / self.data.df_time["delta_t_overhead"]
                    * 100,
                    name="Post runtime",
                    textposition="inside",
                    textangle=0,
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
            self.f_t[-1].update_layout(
                xaxis_title_text="Ranks",
                yaxis_title_text="Percentage (%)",
                width=self.width + self.width_increase,
                height=self.height,
                title="Overhead",
                barmode="stack",
            )  # ,uniformtext_minsize=12, uniformtext_mode='show')
            self.f_t[-1].update_yaxes(range=[0, 100])
            self.f_t[-1] = format_plot(self.f_t[-1])

        if "delta_t_rank0" in self.data.df_time.head():
            self.f_t.append(go.Figure())
            delmiter = "<br>"
            if self.nRun != 1:
                delmiter = " "
                self.f_t[-1].update_layout(barmode="relative")
                self.f_t[-1].update_layout(
                    legend=dict(
                        orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                    )
                )
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_rank0_app"]
                    / self.data.df_time["delta_t_rank0"]
                    * 100,
                    text=self.data.df_time["delta_t_rank0_app"]
                    / self.data.df_time["delta_t_rank0"]
                    * 100,
                    name="App",
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[1],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_rank0_overhead_peri_runtime"]
                    / self.data.df_time["delta_t_rank0"]
                    * 100,
                    text=self.data.df_time["delta_t_rank0_overhead_peri_runtime"]
                    / self.data.df_time["delta_t_rank0"]
                    * 100,
                    name="Overhead%speri-run" % delmiter,
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[7],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
            self.f_t[-1].add_trace(
                go.Bar(
                    x=x,
                    y=self.data.df_time["delta_t_rank0_overhead_post_runtime"]
                    / self.data.df_time["delta_t_rank0"]
                    * 100,
                    text=self.data.df_time["delta_t_rank0_overhead_post_runtime"]
                    / self.data.df_time["delta_t_rank0"]
                    * 100,
                    name="Overhead%spost-run" % delmiter,
                    textposition="inside",
                    textangle=0,
                    marker_color=colors[8],
                    texttemplate=self.barprecision,
                    textfont=dict(color="white"),
                )
            )
            self.f_t[-1].update_layout(
                xaxis_title_text="Ranks",
                yaxis_title_text="Percentage (%)",
                width=self.width + self.width_increase,
                height=self.height,
                title="Overhead Rank 0",
                barmode="stack",
            )
            self.f_t[-1].update_yaxes(range=[0, 100])
            self.f_t[-1] = format_plot(self.f_t[-1])

            self.f_t.append(add_fig_col(self.nRun))
            for i in range(0, self.nRun):
                index = self.data.df_time["file_index"].isin([i])
                self.f_t[-1].add_trace(
                    go.Scatter(
                        x=self.data.df_time["number_of_ranks"][index],
                        y=self.data.df_time["delta_t_rank0"][index],
                        mode="lines+markers",
                        fill="tozeroy",
                        name="Total",
                        legendgroup="Total",
                        marker=dict(
                            symbol=symbols[0], line=dict(width=1, color=markeredgecolor)
                        ),
                        marker_color=colors[0],
                        showlegend=legend_fix(self, i),
                    ),
                    row=1,
                    col=i + 1,
                )
                self.f_t[-1].add_trace(
                    go.Scatter(
                        x=self.data.df_time["number_of_ranks"][index],
                        y=self.data.df_time["delta_t_rank0_app"][index],
                        mode="lines+markers",
                        fill="tozeroy",
                        name="App",
                        legendgroup="App",
                        marker=dict(
                            symbol=symbols[1], line=dict(width=1, color=markeredgecolor)
                        ),
                        marker_color=colors[1],
                        showlegend=legend_fix(self, i),
                    ),
                    row=1,
                    col=i + 1,
                )
                self.f_t[-1].add_trace(
                    go.Scatter(
                        x=self.data.df_time["number_of_ranks"][index],
                        y=self.data.df_time["delta_t_rank0_overhead_post_runtime"][
                            index
                        ]
                        + self.data.df_time["delta_t_rank0_overhead_peri_runtime"][
                            index
                        ],
                        mode="lines+markers",
                        fill="tozeroy",
                        name="Overhead",
                        legendgroup="Overhead",
                        marker=dict(
                            symbol=symbols[2], line=dict(width=1, color=markeredgecolor)
                        ),
                        marker_color=colors[2],
                        showlegend=legend_fix(self, i),
                    ),
                    row=1,
                    col=i + 1,
                )
                self.f_t[-1].update_xaxes(title_text="Ranks", row=1, col=i + 1)
            self.f_t[-1].update_layout(
                xaxis_title="Ranks",
                yaxis_title="Time (s)",
                width=self.width + self.width_increase,
                height=self.height,
                title="Time Distribution Rank 0",
            )
            self.f_t[-1].update_layout(hovermode="x", legend_tracegroupgap=1)
            self.f_t[-1].update_xaxes(
                showspikes=True, spikethickness=1, spikecolor="black"
            )
            self.f_t[-1].update_yaxes(
                showspikes=True, spikethickness=1, spikecolor="black"
            )
            self.f_t[-1] = format_plot(self.f_t[-1])

            modes = [
                "delta_t_rank0",
                "delta_t_rank0_app",
                "delta_t_rank0_overhead_post_runtime + delta_t_rank0_overhead_peri_runtime",
            ]
            names = ["Total", "App", "Overhead"]
            self.f_t.append(
                plot_time_error_bar(
                    self.data.df_time, modes, names, colors, symbols, markeredgecolor
                )
            )

        print(" '-> Finished generating plot I/O time")
        # print('%i figures'%len(self.f))

    # **********************************************************************
    # *                       6. Set and Get Functions
    # **********************************************************************

    def get_figure(self, mode):
        if "write_async" in mode:
            return self.f_aw
        if "read_async" in mode:
            return self.f_ar
        if "write_sync" in mode:
            return self.f_sw
        if "read_sync" in mode:
            return self.f_sr
        if "time" in mode:
            return self.f_t

    def set_figure(self, f, mode):
        # assign
        if "async" in mode:
            if "write" in mode:
                self.f_aw = f
            else:  # read
                self.f_ar = f
        else:  # sync
            if "write" in mode:
                self.f_sw = f
            else:  # read
                self.f_sr = f



from threading import Thread
import numpy as np



class overlap_thread(Thread):
    # constructor
    def __init__(self, ts, te, b_rank_sum, b_rank_avr = []):
        # execute the base constructor
        Thread.__init__(self)
        # set out values
        self.n_overlap     = []
        self.b_overlap_sum = []
        self.b_overlap_avr = []
        self.t_overlap     = []
	
        #set in values
        self.ts = ts
        self.te = te
        self.b_rank_sum = b_rank_sum
        self.b_rank_avr = b_rank_avr
 
    # function executed in a new thread
    def run(self):
        id_s = np.argsort(self.ts)
        id_e = np.argsort(self.te)
        agg_phases  = len(self.ts)
        b_ovr_sum   = []
        b_ovr_avg   = []
        t           = []
        n           = []
        n_tmp       = 0
        b_tmp       = 0
        b2_tmp      = 0
        k_s         = 0
        k_e         = 0
        while (k_s < agg_phases or k_e < agg_phases):
            if (k_s == agg_phases or self.te[id_e[k_e]] < self.ts[id_s[k_s]]):
                t.append(self.te[id_e[k_e]])
                if self.b_rank_sum:
                    b_tmp = b_tmp - self.b_rank_sum[id_e[k_e]] 
                if self.b_rank_avr:
                    b2_tmp = b2_tmp -self.b_rank_avr[id_e[k_e]] 
                n_tmp -=1 
                k_e +=1 
            else:
                t.append(self.ts[id_s[k_s]])
                if self.b_rank_sum:
                    b_tmp = b_tmp + self.b_rank_sum[id_s[k_s]] 
                if self.b_rank_avr:
                    b2_tmp = b2_tmp +self.b_rank_avr[id_s[k_s]] 
                n_tmp +=1 
                k_s +=1
            b_ovr_sum.append(b_tmp)
            b_ovr_avg.append(b2_tmp)
            n.append(n_tmp)
        if self.b_rank_sum:
            self.b_overlap_sum.extend(b_ovr_sum)
            # self.b_overlap_sum.extend(np.zeros(len(b_ovr_avg)))
        if self.b_rank_avr:
            self.b_overlap_avr.extend(b_ovr_avg)
        self. t_overlap.extend(t)
        self. n_overlap.extend(n)


import sys
import time
import os
import subprocess
from multiprocessing import Manager
from rich.console import Console
import numpy as np
import zmq
from ftio.prediction.helper import print_data#, export_extrap
from ftio.prediction.async_process import handle_in_process
from ftio.prediction.probability_analysis import find_probability
from ftio.prediction.helper import get_dominant_and_conf, get_hits
from ftio.api.gekkoFs.ftio_gekko import run
from ftio.prediction.analysis import display_result, save_data, data_analysis
from ftio.prediction.async_process import join_procs
from ftio.freq.helper import MyConsole
from ftio.parse.args import parse_args


T_S = time.time()
CONSOLE = MyConsole()
CONSOLE.set(True)
CARGO = True
CARGO_PATH = "/lustre/project/nhr-admire/vef/cargo/build/cli"
CARGO_STAGE_OUT_PATH ="/lustre/project/nhr-admire/tarraf/stage-out"
CARGO_SERVER = "ofi+sockets://127.0.0.1:62000"


def main(args: list[str] = sys.argv) -> None:
    #parse arguments
    tmp_args = parse_args(args)
    # addr = "10.81.3.98"    
    # port = "5555"
    addr = tmp_args.zmq_address
    port = tmp_args.zmq_port

    if CARGO:
        call = f"{CARGO_PATH}/cargo_ftio --server {CARGO_SERVER}"
        CONSOLE.print("[bold green][Init][/][green]" +call+"\n")
        os.system(call)

        #input is relative from GekokFS
        call = f"{CARGO_PATH}/ccp --server {CARGO_SERVER} --input / --output {CARGO_STAGE_OUT_PATH} --if gekkofs --of parallel"
        CONSOLE.print("[bold green][Init][/][green]" +call+"\n")
        os.system(call)

    ranks = 0
    args.extend(["-e", "plotly", "-f", "10", "-m", "write"])
    context = zmq.Context()
    socket = context.socket(socket_type=zmq.PULL)

    
    try:
        socket.bind(f"tcp://{addr}:{port}")
    except zmq.error.ZMQError as e:
        CONSOLE.print(f"[yellow]Error encountered:\n{e}[/]")
        CONSOLE.print("[yellow]Wrong ip address. FTIO is running on:[/]")
        addr = str(subprocess.check_output("ip addr | grep 'inet 10' | awk  '{print $2}'", shell=True))
        end   = addr.rfind("/")
        start = addr.find("'")
        addr = addr[start+1:end]
        CONSOLE.print("[bold green]correcting Listing IP address[/]")
        socket.bind(f"tcp://{addr}:{port}")
    
    CONSOLE.print(f"[green]FTIO is running on: {addr}[/]")
    # can be extended to listen to multiple sockets
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    # Init
    manager = Manager()
    queue = manager.Queue()
    data = manager.list()  # stores prediction
    aggregated_bytes = manager.Value("d", 0.0)
    hits = manager.Value("d", 0.0)
    start_time = manager.Value("d", 0.0)
    count = manager.Value("i", 0)
    procs = []
    b_app = manager.list()
    t_app = manager.list()

    # for Cargo trigger process:
    sync_trigger = manager.Queue()
    trigger = handle_in_process(trigger_cargo, args=(sync_trigger,)) 

    if "-zmq" not in args:
        args.extend(["--zmq"])

    # Loop and predict if changes occur
    try:
        with CONSOLE.status("[green] started\n",spinner="arrow3") as status:
            while True:
                if procs:
                    procs = join_procs(procs)

                # get all messages
                msgs = []
                ranks = 0
                socks = dict(poller.poll(1000))
                start = time.time()
                while socks and time.time() < start + 0.5:
                    if socks.get(socket) == zmq.POLLIN:
                        msg = socket.recv(zmq.NOBLOCK)
                        msgs.append(msg)
                        # CONSOLE.print(f"[cyan]Got message {ranks}:[/] {msg}")
                        ranks += 1
                    socks = dict(poller.poll(1000))

                if not msgs:
                    # CONSOLE.print("[red]No messages[/]")
                    status.update("[cyan]Waiting for messages\n",spinner="dots")
                    continue
                CONSOLE.print(f"[cyan]Got message from {ranks}:[/]")
                status.update("")

                # launch prediction_process
                procs.append(
                    handle_in_process(
                        prediction_zmq_process,
                        args=(
                            data,
                            queue,
                            count,
                            hits,
                            start_time,
                            aggregated_bytes,
                            args,
                            msgs,
                            b_app,
                            t_app,
                            sync_trigger
                        ),
                    )
                )

    except KeyboardInterrupt:
        trigger.join()
        print_data(data)
        # export_extrap(data=data)
        print("-- done -- ")


def prediction_zmq_process(
    data,
    queue,
    count,
    hits,
    start_time,
    aggregated_bytes,
    args: list[str],
    msg,
    b_app,
    t_app,
    sync_trigger
) -> None:
    """performs prediction

    Args:
        data (_type_): _description_
        queue (_type_): _description_
        count (_type_): _description_
        hits (_type_): _description_
        start_time (_type_): _description_
        aggregated_bytes (_type_): _description_
        args (list[str]): _description_
        msg (_type_): _description_
        b_app (_type_): _description_
        t_app (_type_): _description_
        sync_trigger (_type_): _description_
    """
    t_prediction = time.time()
    console = Console()
    console.print(f"[purple][PREDICTOR] (#{count.value}):[/]  Started")

    # Modify the arguments
    args.extend(["-e", "no"])
    args.extend(["-ts", f"{start_time.value:.2f}"])

    # Perform prediction
    prediction, args, t_flush = run(msg, args, b_app, t_app)

    # get data
    freq, conf = get_dominant_and_conf(prediction)  # just get a single dominant value
    hits = get_hits(prediction, count.value, hits)

    # save prediction results
    save_data(queue, prediction, aggregated_bytes, count, hits)
    # display results
    text = display_result(freq, prediction, count, aggregated_bytes)
    # data analysis to decrease window
    text, start_time.value = data_analysis(args, prediction, freq, count, hits, text)
    console.print(text)
    count.value += 1

    # save data to queue
    while not queue.empty():
        data.append(queue.get())

    #calculate probability
    prob = find_probability(data)

    probability = -1
    for p in prob:
        if p.get_freq_prob(freq):
            probability = p.p_freq_given_periodic
            break

    # send data to trigger proc
    sync_trigger.put(
        {
    't_wait':  time.time() ,
    't_end':  prediction['t_end'],
    't_start':  prediction['t_start'],
    't_flush': t_flush + (t_prediction- time.time()),
    'freq': freq,
    'conf': conf,
    'probability': probability
        })
        
    CONSOLE.print("[bold purple]Ended[/]")



def trigger_cargo(sync_trigger):
    """sends cargo calls. For that in extracts the predictions from `sync_trigger` and examines it. 

    Args:
        sync_trigger (_type_): _description_
    """
    while True:
        try:
            if not sync_trigger.empty():
                skip_flag = False 
                prediction = sync_trigger.get()
                t = time.time() - prediction['t_wait']  # add this time to t_flush (i.e., the time waiting)
                # CONSOLE.print(f"[bold green][Trigger] queue wait time = {t:.3f} s[/]\n")
                if not np.isnan(prediction['freq']):
                    #? 1) Find estimated number of phases and skip in case less than 1
                    # n_phases = (prediction['t_end']- prediction['t_start'])*prediction['freq']
                    # if n_phases <= 1:
                    #     CONSOLE.print(f"[bold green][Trigger] Skipping this prediction[/]\n")
                    #     continue
                    
                    #? 2) Time analysis to find the right instance when to send the data
                    target_time = prediction['t_end'] + 1/prediction['freq']
                    geko_elapsed_time = prediction['t_flush'] + t  # t  is the waiting time in this function
                    remaining_time = (target_time - geko_elapsed_time ) 
                    CONSOLE.print(f"[bold green][Trigger][/][green] Target time = {target_time:.3f} -- Gekko time = {geko_elapsed_time:.3f} -> sending cmd in {remaining_time:.3f} s[/]\n")
                    if remaining_time > 0:
                        countdown = time.time() + remaining_time
                        # wait till the time elapses:
                        while time.time() < countdown:
                            pass
                            #? 3) Skip in case new prediction is available  
                            # if not sync_trigger.empty():
                            #     skip_flag = True

                        if not skip_flag:
                            if CARGO:
                                # call = f"{CARGO_PATH}/cargo_ftio --server {CARGO_SERVER} -c {prediction['conf']} -p {prediction['probability']} -t {1/prediction['freq']}"
                                call = f"{CARGO_PATH}/cargo_ftio --server {CARGO_SERVER} --run"
                                os.system(call)

                            CONSOLE.print("[bold green][Trigger][/][green]" +call+"\n")
                        else:
                            CONSOLE.print("[bold green][Trigger][/][yellow] Skipping, new prediction is ready[/]\n")

                    else:
                        CONSOLE.print("[bold green][Trigger][/][yellow] Skipping, not in time[/]\n")

            time.sleep(0.01)
        except KeyboardInterrupt:
            exit()


if __name__ == "__main__":
    main(sys.argv)



"""Module concerned with probability calculation
"""
from __future__ import annotations
import numpy as np
from rich.console import Console


CONSOLE = Console()

class Probability():
    """Class that stores the conditional probability according to frequency intervals
    """
    def __init__(self, freq_min, freq_max, p_periodic = 0, p_freq = 0, p_freq_given_periodic = 0, p_periodic_given_freq=1):
        """init function

        Args:
            freq_min (float): minimum frequencies in group
            freq_max (float): max frequencies in group
            p_periodic (float, optional): probability of being periodic. Defaults to 0.
            p_freq (float, optional): Probability that the signal has the frequency in the range [freq_min, freq_max]. Defaults to 0.
            p_freq_given_periodic (float, optional): conditional probability that expresses how probable the frequency is in the interval [freq_min, freq_max] ,
            given that the signal is periodic. Defaults to 0.
            p_periodic_given_freq (float, optional): Probability that the signal is periodic given that it has a frequency A. Defaults to 1.
        """
        self.freq_min = freq_min
        self.freq_max = freq_max
        self.p_periodic = p_periodic
        self.p_freq =  p_freq
        self.p_freq_given_periodic = p_freq_given_periodic
        self.p_periodic_given_freq = p_periodic_given_freq

    def set(self, p_periodic = np.nan, p_freq = np.nan, p_freq_given_periodic = np.nan, p_periodic_given_freq=np.nan):
        self.p_periodic = p_periodic if not np.isnan(p_periodic) else self.p_periodic
        self.p_freq =  p_freq if not np.isnan(p_freq) else self.p_freq
        self.p_freq_given_periodic = p_freq_given_periodic if not np.isnan(p_freq_given_periodic) else self.p_freq_given_periodic
        self.p_periodic_given_freq = p_periodic_given_freq if not np.isnan(p_periodic_given_freq) else self.p_periodic_given_freq

    def display(self):
        CONSOLE.print(
                    f"[purple][PREDICTOR]:[/] P([{self.freq_min:.3f},{self.freq_max:.3f}] Hz) = {self.p_periodic*100:.2f}%\n"
                    f"[purple][PREDICTOR]:[/] |-> [{self.freq_min:.3f},{self.freq_max:.3f}] Hz = [{1/self.freq_max if self.freq_max != 0 else np.NaN:.3f},{1/self.freq_min if self.freq_min != 0 else np.NaN:.3f}] sec\n"
                    f"[purple][PREDICTOR]:[/] '-> P([{self.freq_min:.3f},{self.freq_max:.3f}] Hz | periodic) = {self.p_freq_given_periodic*100:.2f}%\n"
                )

    def get_freq_prob(self, freq):
        if freq >= self.freq_min and freq <= self.freq_max:
            return True
        else: 
            return False





import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from ftio.freq.helper import format_plot


def label_phases(
    prediction: dict, args, b0: np.ndarray = np.array([]), t0: np.ndarray = np.array([])
):
    """Labels the phases using the result from FTIO

    Args:
        prediction (dict): prediction from FTIO
        args (Argparse): arguments passed to FTIO
        b0 (np.ndarray, optional): original bandwidth. Defaults to np.array([]).
        t0 (np.ndarray, optional): original time. Defaults to np.array([]).

    Returns:
        _type_: _description_
    """
    # just take the dominant or consider more than one
    only0ne = False
    n_waves = 1

    t = np.arange(prediction["t_start"], prediction["t_end"], 1 / prediction["freq"])

    if only0ne:
        dominant_index = np.argmax(prediction["conf"])
        # conf = prediction["conf"][dominant_index]
        f = prediction["dominant_freq"][dominant_index]
        amp = prediction["amp"][dominant_index]
        phi = prediction["phi"][dominant_index]
        n = np.floor((prediction["t_end"] - prediction["t_start"]) * prediction["freq"])

        ## create cosine wave
        cosine_wave = 2 * amp / n * np.cos(2 * np.pi * f * t + phi)
    else:
        cosine_wave = np.zeros(len(t))
        n_waves = len(prediction["conf"]) if args.n_freq == 0 else len(prediction["top_freq"]["freq"])
        print(f"merging {n_waves} frequencies")

        # iterate over all dominant frequencies or take the top_frequencies if args.n_freq is above 0
        for index, _ in enumerate(
            prediction["conf"] if args.n_freq == 0 else prediction["top_freq"]["freq"]
        ):
            # conf = prediction["conf"][index]
            f = (
                prediction["dominant_freq"][index]
                if args.n_freq == 0
                else prediction["top_freq"]["freq"][index]
            )
            amp = (
                prediction["amp"][index]
                if args.n_freq == 0
                else prediction["top_freq"]["amp"][index]
            )
            phi = (
                prediction["phi"][index]
                if args.n_freq == 0
                else prediction["top_freq"]["phi"][index]
            )
            n = np.floor(
                (prediction["t_end"] - prediction["t_start"]) * prediction["freq"]
            )

            #skip frequency at 0
            if f == 0 and args.n_freq != 0:
                continue
            
            ## create cosine wave
            cosine_wave = cosine_wave + 2 * amp / n * np.cos(2 * np.pi * f * t + phi)

    ## make square signal
    square_wave = np.zeros(len(cosine_wave))
    top = np.max(cosine_wave)
    for i, x in enumerate(cosine_wave):
        if x > 0:
            square_wave[i] = top
        else:
            square_wave[i] = -top

    phases = []
    tmp = []
    for i, x in enumerate(square_wave):
        if i == 0:
            counter = 0
            mem = x
            phases.append(counter)
            tmp.append(t[i])
        else:
            if mem == x:
                pass
            else:
                mem = x
                counter += 1
                phases.append(counter)
                tmp.append(t[i])

    # Assign in not empty
    if tmp:
        # add end time
        tmp.append(t[-1])
        time = {"t_s": tmp[0:-1], "t_e": tmp[1:]}
    else:
        time = {"t_s": [], "t_e": []}

    ## Plot
    _ = plot_classification(args, cosine_wave, t, time, n_waves, b0, t0)

    return phases, time


def plot_classification(args, cosine_wave, t, time, n_waves = 0, b0 =np.array([]), t0=np.array([])) -> go.Figure | None:
    fig = None
    if any(x in args.engine for x in ["mat", "plot"]):
        fig = go.Figure()
        if n_waves == 1:
            name = "Dominant wave" 
        elif args.n_freq == 0:
            name = "Dominant waves"
        else:
            name = f"{n_waves} superposed <br>cosine waves"
        fig.add_trace(
            go.Scatter(x=t, y=cosine_wave, name=name, marker_color="red")
        )
        # fig.add_trace(go.Scatter(x=t, y=square_wave, name="square wave"))
        fig.add_hline(y=0, line_width=1, line_color="gray")
        colors = px.colors.qualitative.Plotly + px.colors.qualitative.G10
        for i, _ in enumerate(time["t_s"]):
            if i == len(colors):
                print("Not enough colors")
                break
            fig.add_vrect(
                x0=time["t_s"][i],
                x1=time["t_e"][i],
                annotation_text=f"phase {i}",
                annotation_position="top",
                fillcolor=colors[i],
                opacity=0.25,
                line_width=0,
            )

        if b0.size > 0:
            fig.add_trace(
                go.Scatter(
                    x=t0,
                    y=b0,
                    mode="lines+markers",
                    name="Original signal",
                    marker_color="rgb(0,150,250)",
                )
            )

        fig.update_layout(
            xaxis_title="Ranks",
            yaxis_title="Transfer Rate (B/s)",
            width=1800,
            height=600,
        )
        fig = format_plot(fig)
        fig.show()

        return fig



class FreqData:
    """contains the data from the frequency analysis including the 
    1. bandwidth over time spaced with constant steps (1/freq)
    2. settings
    3. original data before sampling
    """
    def __init__(self, df1, df2, df3):
        self.data_df      = df1
        self.settings_df  = df2
        self.original_df  = df3



"""Outlier plot function
"""

from __future__ import annotations
import numpy as np
import plotly.graph_objects as go
import plotly.express as px
import matplotlib.pyplot as plt
from sklearn.inspection import DecisionBoundaryDisplay
from ftio.freq.freq_html import create_html
from ftio.freq.helper import format_plot
from ftio.freq.freq_plot_core import plot_both_spectrums


# ?#################################
# ? Plot outliers
# ?#################################
def plot_outliers(
    args,
    freq_arr: np.ndarray,
    amp: np.ndarray,
    indecies: np.ndarray,
    conf: np.ndarray,
    dominant_index: np.ndarray,
    d: np.ndarray = np.array([]),
    eps: float = 0,
) -> None:
    """Plots outliers for DB scan

    Args:
        freq_arr (np.ndarray): _description_
        amp (np.ndarray): aplitude or power
        indecies (np.ndarray): _description_
        conf (np.ndarray): _description_
        dominant_index (np.ndarray): _description_
        d (np.ndarray, optional): _description_. Defaults to np.array([]).
    """
    colorscale = [(0, "rgb(0,50,150)"), (0.5, "rgb(150,50,150)"), (1, "rgb(255,50,0)")]
    labels = [(f"cluster {i}") if (i >= 0) else "outliers" for i in dominant_index]
    #prepare figures
    figs = []
    for i in np.arange(0, 5):
        f = go.Figure()
        f = format_plot(f)
        figs.append(f)

    #prepare symbols
    if dominant_index.size > 0 and dominant_index.max() < 20:
        symbol = dominant_index.copy()
        symbol[symbol >= 5] = symbol[symbol >= 5] + 1
        symbol[symbol == -1] = 5  # x marker
        symbol = symbol - 1
    else:
        symbol = np.ones(dominant_index.size)
    labels = np.array(labels)

    if d.size  == 0 and len(freq_arr) != 0:
        d = np.vstack(
            (
                freq_arr[indecies] / freq_arr[indecies].max(),
                amp[indecies] / amp[indecies].sum(),
            )
        ).T
    elif len(freq_arr) == 0:
        return
    else:
        pass

    
    all_colors = [px.colors.qualitative.Alphabet[0], px.colors.qualitative.Plotly[0]] + px.colors.qualitative.Plotly[2:]
    all_colors = np.array(all_colors + px.colors.qualitative.Alphabet + px.colors.qualitative.Dark24 + [px.colors.qualitative.Plotly[1]])

    # Only for DB SCAN
    if args.outlier.lower() in ["dbscan", "db-scan", "db"]:
        if dominant_index.max() < 20: 
            color = all_colors[dominant_index]
        else:
            color = np.array([
                        "blue" if (x >= 0) else "red"
                        for x in dominant_index
                    ])
        # draw the circles
        for i in range(0, len(d)):
            figs[0].add_shape(
                dict(
                    type="circle",
                    x0=d[i, 0] - eps,
                    y0=d[i, 1] - eps,
                    x1=d[i, 0] + eps,
                    y1=d[i, 1] + eps,
                    opacity=0.3,
                ),
                name=labels[i],
                line_color=color[i],
            )
    else:     
        color = all_colors[dominant_index]

    for i in np.unique(labels):
        figs[0].add_trace(
            go.Scatter(
                x=d[labels == i, 0],
                y=d[labels == i, 1],
                mode="markers",
                marker=dict(color=color[labels == i], colorscale=colorscale),
                text=conf[labels == i],
                hovertemplate="<b>Noremed values<br><br>Freq: %{x:.4f}  <br>"
                + "Amplitude: %{y:.2f}<br>"
                + "conf: %{text}",
                showlegend=True,
                name=i,
            )
        )

        figs[1].add_trace(
            go.Scatter(
                x=d[labels == i, 0],
                y=d[labels == i, 1],
                mode="markers",
                marker=dict(
                    color=conf[labels == i],
                    colorscale=colorscale,
                    coloraxis="coloraxis",
                    symbol=symbol[labels == i],
                    size=12,
                ),
                text=labels[labels == i],
                hovertemplate="<b>Noremed values<br><br>Freq: %{x:.4f}  <br>"
                + "Amplitude: %{y:.2f}<br>"
                + "cluster: %{text}",
                showlegend=True,
                name=i,
            )
        )

        figs[2].add_trace(
            go.Scatter(
                x=freq_arr[indecies[labels == i]],
                y=2 * amp[indecies[labels == i]],
                mode="markers",
                marker=dict(color=color[labels == i], colorscale=colorscale),
                text=labels[labels == i],
                hovertemplate="<b>Freq: %{x:.4f}    Hz<br>"
                + "Amplitude: %{y:.2f}<br>"
                + "cluster: %{text}",
                showlegend=True,
                name=i,
            )
        )

    figs[1].update_layout(
        coloraxis={
            "colorbar": {
                "x": 0.92,
                "len": 0.91,
                "y": 0.57,
                "title": "conf",
                "thickness": 10,
                "tickfont": dict(size=14),
            },
            "colorscale": colorscale,
        }
    )
    
    counter = 2
    spec_figs, plt_names = plot_both_spectrums(args, freq_arr, amp, full=False)
    for trace in list(spec_figs.select_traces()):
        counter += 1
        trace.update(marker={"coloraxis": "coloraxis"})
        figs[counter].add_trace(trace)
        if figs[counter].data and "xaxis" in figs[counter].data[0]:
            figs[counter].data[0]["xaxis"] = "x1"
            figs[counter].data[0]["yaxis"] = "y1"

    for i in np.arange(3, 5):
        figs[i].update_layout(coloraxis={"colorscale": colorscale})

    y_title = "Power" if args.psd else "Amplitude"
    figs[0].update_xaxes(title_text="Normed Frequency", range=[-0.01, 1.01])
    figs[0].update_yaxes(title_text=f"Normed {y_title}")
    figs[1].update_xaxes(title_text="Normed Frequency ", range=[-0.01, 1.01])
    figs[1].update_yaxes(title_text=f"Normed {y_title}")
    figs[2].update_xaxes(title_text="Frequency (Hz)",range=[0, freq_arr[indecies].max()])
    figs[2].update_yaxes(title_text=f"{y_title}")
    figs[3].update_xaxes(title_text="Frequency (Hz)")
    figs[3].update_yaxes(title_text=plt_names[0])
    figs[4].update_xaxes(title_text="Frequency (Hz)")
    figs[4].update_yaxes(title_text=plt_names[1])
    for fig in figs:
        fig.update_layout(width=1300, height=400)
    configuration = {"toImageButtonOptions": {"format": "png", "scale": 4}}
    create_html(figs, args.render, configuration, "anaomality")


def plot_decision_boundaries(model, d, conf):
    disp = DecisionBoundaryDisplay.from_estimator(
        model,
        d,
        response_method="decision_function",
        alpha=0.5,
    )
    disp.ax_.scatter(d[:, 0], d[:, 1], c=conf, s=20, edgecolor="k")
    disp.ax_.set_title("Binary decision boundary \nof IsolationForest")
    plt.axis("square")
    plt.legend(labels=["outliers", "inliers"], title="true class")
    plt.show()




from pathlib import Path
import json
import msgpack
import numpy as np

def parse(file_path_or_msg, data, io_type = "w") -> tuple[dict, str]:
    """Parses data from gekko

    Args:
        file_path_or_msg (_type_): list of files or messages (ZMQ)
        data (_type_): data to append to 
        io_type (str, optional): Can be w for write or r for read. Defaults to "w".

    Raises:
        RuntimeError: _description_

    Returns:
        tuple[dict, str]: _description_
    """
    if isinstance(file_path_or_msg,bytes):
        extension = "ZMQ"
    else:
        extension = Path(file_path_or_msg).suffix

    # ZMQ
    if "ZMQ" in extension.upper():
        # if data is no struct:
        # unpacked_data = msgpack.unpackb(file_path_or_msg)
        # else:
        unpacker = msgpack.Unpacker()
        unpacker.feed(file_path_or_msg)
        data = assign(data, unpacker)

    # MsgPack
    elif "MSG" in extension.upper():
        # Read the binary data
        with open(file_path_or_msg, "rb") as in_file:
            binary_data = in_file.read()
            
        # Deserialize the MessagePack data 
        unpacker = msgpack.Unpacker()
        unpacker.feed(binary_data)
        data = assign(data, unpacker, io_type)

    # JSON
    elif "JSON" in extension.upper(): 
        with open(file_path_or_msg, 'r') as json_file:
                json_data = json.load(json_file)
        for key, value in json_data.items():
            if 'avg_thruput_mib' in key :
                data['avg_thruput_mib'].extend(value)
            elif 'start_t_micro' in key :
                data['start_t_micro'].extend(value)
            elif 'end_t_micro' in key :
                data['end_t_micro'].extend(value)
            elif 'req_size' in key :
                data['req_size'].extend(value)
            elif 'hostname' in key :
                data['hostname'] = value
            elif 'flush_t' in key :
                data['flush_t'] = value
            elif 'pid' in key: 
                data['pid'] = value
            elif 'total_bytes' in key :
                data['total_bytes'] += value
            elif 'total_iops' in key :
                data['total_iops'] += value
            elif 'io_type' in key :
                data['io_type'] += value

    else:
        raise RuntimeError("Unsupported file format specified")

    return data, extension


def assign(data:dict,unpacker, io_type = "w")->dict:
    data_fields = ["flush_t", "hostname", "pid", "io_type","start_t_micro", "end_t_micro", "req_size", "total_iops","total_bytes"]
    index = 0
    skip = False
    
    for item in unpacker:
        if isinstance(item,dict):
            item = item[data_fields[index]]
        if index in [4,5,6]:
            data[data_fields[index]].extend(item)
        elif index == 0: #find max flush time
            data[data_fields[index]] = max(data[data_fields[index]],item)
        else:
            data[data_fields[index]] = item
            # exit if it is not the right mode
            if index == 3 and item != io_type:
                skip = True
                break
        index += 1

    if not skip:
        b = np.array(object=data["req_size"])/(np.array(data["end_t_micro"]) - np.array(data["start_t_micro"]))
        if np.isnan(b).any() or np.isnan(b).any():
            print(f'b_rank : {b} \nt_rank_s : {data["start_t_micro"]} \nt_rank_e : {data["end_t_micro"]} \n')
            b[b==np.nan] = 0
            b[b==np.inf] = 0

        data['avg_thruput_mib'].extend(b)
    else:
        # print("Skipping ... ")
        pass

    return data



"""Performs action async to current process
"""
from __future__ import annotations
from typing import Callable
from multiprocessing import Process

def handle_in_process(function: Callable, args) -> Process:
    """Handle function in a dedicated process

    Args:
        function (Callable): function name
        args (argparse): arguments passed to function

    Returns:
        None
    """
    process = Process(target=function, args=args) 
    # print(f'Process {process.name} (PID {os.getpid()}) started to execute {function}')
    process.start()
    # print(f'Process {process.name} (PID {os.getpid()}) ended')
    # print(f"Process {process} created")
    return process

def join_procs(procs:list) -> list:
    """joins procs by iterating over list and testing
    that the process finished before joining

    Args:
        procs (list): list of procs to test

    Returns:
        list of new procs
    """
    if procs:
        for p in procs:
            if p.is_alive():
                pass 
                # print(f"Process {p} still working")
            else:
                p.join()
                procs.remove(p)
                # print(f"Process {p} JOINED")
    return procs


'''Parse zmq message containing three fields: 
the bandwidth, the start time, and the end time.
'''
import zmq
from ftio.parse.simrun import Simrun
from ftio.parse.zmq_reader import extract
from ftio.parse.msgpack_reader import extract_data
from ftio.freq.helper import MyConsole

class ParseZmq:
    '''class to parse zmq file
    '''
    def __init__(self, msgs):
        self.msgs = msgs

    def to_simrun(self, args, index:int = 0):
        '''Convert to Simrun class
        Args:
            ars (argparse): command line arguments
            index: file index in case several files are passed
        Returns:
            Simrun: Simrun object
        '''
        # For FTIO not PREDICTOR
        if self.msgs is None:
            self.msgs = get_msgs_zmq(args)

            
        if 'direct' in args.zmq_source:
            for msg in self.msgs:
                dataframe, ranks = extract(msg, args)
            return Simrun(dataframe, 'txt',str(ranks), args, index)
        elif 'tmio' in args.zmq_source.lower():
            data = extract_data(self.msgs[0], [])
            return Simrun(data,'msgpack', '0', args, index)
        else:
            pass

# For FTIO only
def get_msgs_zmq(args)-> list[str]:
    CONSOLE = MyConsole()
    CONSOLE.set(True)
    context = zmq.Context()
    socket = context.socket(socket_type=zmq.PULL)

    # socket.bind('tcp://*:5555')
    socket.bind(f'tcp://{args.zmq_address}:{args.zmq_port}')

    # can be extended to listen to multiple sockets
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    msgs = []
    ranks = 0
    socks = dict(poller.poll(1000))
    while(True):
        while(socks):
            if socks.get(socket) == zmq.POLLIN:
                msg = socket.recv(zmq.NOBLOCK)
                msgs.append(msg)
                ranks += 1
            socks = dict(poller.poll(1000))

        if not msgs:
            CONSOLE.print('[red]No messages[/]')
            continue
        else:
            CONSOLE.print('[green]All message received[/]')
            break

    return msgs


"""Performs prediction with Pools (ProcessPoolExecutor) and a callback mechanism"""
from __future__ import annotations
import zmq
from ftio.prediction.async_process import join_procs
from ftio.prediction.processes import prediction_process
from ftio.prediction.helper import print_data, export_extrap
from ftio.prediction.async_process import handle_in_process
from ftio.parse.args import parse_args
from ftio.freq.helper import MyConsole

CONSOLE = MyConsole()
CONSOLE.set(True)

def predictor_with_processes_zmq(
    data, queue, count, hits, start_time, aggregated_bytes, args, b_app, t_app
)-> None:
    """performs prediction in ProcessPoolExecuter. FTIO is a submitted future and probability is calculated as a callback

    Args:   
        filename (str): name of file
        data (Manager().list): List of dicts with all predictions so far
        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio
    """
    procs = []
    context = zmq.Context()
    socket = context.socket(socket_type=zmq.PULL)
    tmp_args = parse_args(args)
    socket.bind(f'tcp://{tmp_args.zmq_address}:{tmp_args.zmq_port}')
    
    # can be extended to listen to multiple sockets
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    
    if '-zmq' not in args:
        args.extend(['--zmq'])
    
    # Loop and predict if changes occur
    try:
        while True:
            if procs:
                procs = join_procs(procs)
            #1) just a single msg
            # msg = socket.recv(zmq.NOBLOCK)  

            #2) Loop and accept messages from both channels, acting accordingly
            # if socks:
            #     if socks.get(socket) == zmq.POLLIN:
            #         print(f"got message ",{socket.recv(zmq.NOBLOCK)})
            # else:
            #     print("No message received")
            #     continue
            
            #3) Loop and accept messages from both channels, acting accordingly
            msgs = []
            ranks = 0
            socks = dict(poller.poll(1000))
            while(socks):
                if socks.get(socket) == zmq.POLLIN:
                    msg = socket.recv(zmq.NOBLOCK)
                    msgs.append(msg)
                    # CONSOLE.print(f"[cyan]Got message {ranks}:[/] {msg}")
                    ranks += 1
                socks = dict(poller.poll(1000))

            if not msgs:
                CONSOLE.print("[red]No messages[/]")
                continue
            CONSOLE.print(f"[green]All message received from {ranks} ranks[/]")

            # launch prediction_process
            # TODO: append b_app and t_app like predictor_gekko_zmq use the flag --zmq to indicate this
            # put all in msgs and call it zmq_data
            procs.append(
                handle_in_process(
                    prediction_process,
                    args=(data, queue, count, hits, start_time, aggregated_bytes, args, msgs)
                )
            )
    except KeyboardInterrupt:
        print_data(data)
        export_extrap(data)
        print('-- done -- ')





from ftio.parse.simrun import Simrun
from ftio.parse.recorder_reader import extract

class ParseRecorder:
    
    def __init__(self, path):
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1]

    def to_simrun(self, args):
        """Convert to Simrun class
        Args:
            ts (double): [optional] desired start time
        Returns:
            Simrun: Simrun object
        """
        data, ranks = extract(self.path, args)

        return Simrun(data, 'recorder', str(ranks), args)



"""This functions performs the prediction with the the help of ftio.py
"""
from __future__ import annotations
import sys
from multiprocessing import Manager
from ftio.prediction.pools import predictor_with_pools
from ftio.prediction.processes_zmq import predictor_with_processes_zmq
from ftio.prediction.processes import predictor_with_processes

def main(args: list[str] = sys.argv) -> None:
    """runs the prediction and launches new threads whenever data is available

    Args:
        args (list[str]): arguments passed from command line
    """
    # Init
    manager = Manager()
    filename = args[1]
    queue = manager.Queue()
    data = manager.list() # stores prediction
    aggregated_bytes = manager.Value("d", 0.0)
    hits = manager.Value("d", 0.0)
    start_time = manager.Value("d", 0.0)
    count = manager.Value('i', 0)
    b_app = manager.list()
    t_app = manager.list()
    
    mode = "procs" # "procs" or "pool"
    
    if "pool" in mode.lower():
        # prediction with a Pool of process and a callback mechanism
        predictor_with_pools(filename, data, queue, count, hits, start_time, aggregated_bytes, args)
    else:
        if any("zmq" in x for x in args):
            # prediction with Processes of process and a callback mechanism + zmq
            predictor_with_processes_zmq(data, queue, count, hits, start_time, aggregated_bytes, args, b_app, t_app)
        else:
            # prediction with Processes of process and a callback mechanism
            predictor_with_processes(filename, data, queue, count, hits, start_time, aggregated_bytes, args)

if __name__ == "__main__":
    main(sys.argv)



import sys
import re
import time 
from rich.console import Console
from ftio.parse.input_template import init_data
from ftio.parse.custom_patterns import convert

def extract(path:str, args:list, custom:bool = False) -> tuple[dict, int]:
    # init
    start = time.time()
    ranks = 0
    mode, io_data, io_time = init_data(args)
    
    # Define your pattern and translate it
    if not custom:
        pattern, translate = convert()
    else:
        data = {}
        with open(f"{args.custom_file}") as f:
            exec(f.read(),globals(),data)
        pattern = data['pattern']
        translate = data['translate']

    #read data
    extracted_data = read(path, pattern)

    # Display the extracted data
    for key, value in translate.items():
        if "bandwidth" in key:
            for key2, value2 in translate["bandwidth"].items():
                unit, value2 = find_scale(value2)
                io_data["bandwidth"][key2] = extracted_data[value2]
                io_data["bandwidth"][key2] = scale(io_data["bandwidth"][key2], unit)
        else: 
            unit, value = find_scale(value)
            io_data[key] = extracted_data[value]
            io_data[key] = scale(io_data[key],unit)
    
    
    console = Console()
    console.print(f"[cyan]Elapsed time:[/] {time.time()-start:.3f} s")
        
    # fill time
    kind = mode[0] + mode.split('_')[1][0]
    if "a" in mode[0]:
        kind = kind + "a"
    io_time[f"delta_t_{kind}"] = 0
    
    #pack everything
    data = {
        f"{mode}": io_data,
        "io_time": io_time,
    }

    return data, ranks


def read(file_path, patterns):
    results = {}
    with open(file_path, 'r') as file:
        data = file.read()
        for key, pattern in patterns.items():
            match = re.search(pattern, data)
            if match:
                if ',' in match.group(1):
                    # If the matched group contains commas, split and convert to integers
                    tmp = match.group(1).replace(', ',',')
                    results[key] = list(float(val) if '.' in val else int(val) for val in tmp.split(','))

                else:
                    results[key] = int(match.group(1))
            else:
                print(f"Unable to extract data for {key} from the file.")

    return results


def find_scale(value:dict):
    unit = 1
    if isinstance(value,tuple):
        unit = value[1]
        value = value[0]
    return unit,value


def scale(mylist:list, unit:float) -> list:
    """scales lists with unit

    Args:
        mylist (list): input list
        unit (float): unit 

    Returns:
        list: scale
    """
    if unit != 1:
        return [unit * i for i in mylist]
    else: 
        return mylist


def main(args) -> None:
    """Pass varibales and call main_core. The extraction of the traces
    and the parsing of the arguments is done in this function.
    """
    file = args[1]
    data, _ = extract(file, args[1:])
    print(data)


if __name__ == "__main__":
    main(sys.argv)


import time
import msgpack
from rich.console import Console
from ftio.parse.input_template import init_data


def extract(msgs, args:list) -> tuple[dict, int]:
    # init
    start = time.time() 
    ranks = 0
    mode, io_data, io_time = init_data(args)

    unpacked_data = msgpack.unpackb(msgs)

    # Access the data
    ranks           = unpacked_data["ranks"]
    b               = unpacked_data["b"]
    ts              = unpacked_data["ts"]
    te              = unpacked_data["te"]
    # received_float  = unpacked_data["floatData"]
    
    io_data["bandwidth"]["b_rank_avr"] = b
    io_data["bandwidth"]["t_rank_s"]   = ts
    io_data["bandwidth"]["t_rank_e"]   = te

    console = Console()
    console.print(f"[cyan]Elapsed time:[/] {time.time()-start:.3f} s")
    # io_time[f"delta_t_{kind}"] = 0
    
    #pack everything
    data = {
        f"{mode}": io_data,
        "io_time": io_time,
    }

    return data, ranks




from __future__ import annotations
import argparse
from ftio import __copyright__, __repo__, __version__, __license__



def parse_args(argv:list, name='') -> argparse.Namespace:
    flag = True
    if name == '':
        name = argv[0]
        name = name[name.rfind('/')+1:]
    else:
        #API call
        flag = False
        

    if 'plot' in name:
        disc = 'Plots result stored in Json file to a HTML page or PDF document.'
    elif 'ftio' in name:
        disc = 'Captures the period of the I/O phases. Uses frequency techniques (default=discrete fourier transformation) and outlier detection methods (Z-score) on the provided file. Supported file formats are Json, Jsonlines, Msgpack, Darshan, and REcorder (folder). TMIO can be used to generate the tracing file needed. There are several parameters which can be controlled by the arguments bellow.'
    elif 'predictor' in name:
        disc = 'Wrapper code to execute ftio online. Monitors a file for changes. Whenever the file is modified (i.e., new traces are appended) a new prediction process is executed and the result is store in a shared memory space. All parameters that can be passed to ftio are supported by predictor.'
    elif 'parse' in name:
        disc = 'Parses to an extra-p format.'
    else:
        disc = ''

    parser = argparse.ArgumentParser(
        prog=name,#.capitalize(),
        description=disc,
        epilog=f'''
--------------------------------------------
Author: 
Ahmad H. Tarraf

Report any bugs to:
{__repo__}/issues

COPYRIGHT:
{__copyright__} 

LICENCE:
{__license__} 

Full documentation:
{__repo__}
--------------------------------------------
''',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    #! If CLI call not API, add files
    if flag:
        parser.add_argument('files', metavar='files', type = str, nargs='+', help='file, file list (file 0 ... file n), folder, or folder list (folder 0.. folder n)')    

    parser.add_argument('-m', '--mode', dest = 'mode',      type = str ,  help ='if the trace file contains several I/O modes, a specific mode can be selected. Supported modes are: write_async, read_async, write_sync, read_sync')
    parser.add_argument('-s','--source', dest='source', type = str, help = 'the source of the files: tmio, or custom. See https://github.com/tuda-parallel/FTIO/blob/main/docs/file_formats.md')
    parser.set_defaults(source = 'unspecified')
    
    #! PARSE Settings
    if 'parse' not in name.lower():
        parser.add_argument('-r', '--render', dest='render',  type = str ,  help ='specifies how the plots are rendered. Either dynamic (default) or static' )
        parser.set_defaults(render='dynamic')

    #! PLAY Settings
    if 'play' in name.lower():
        parser.add_argument('-f', '--freq', dest='freq', type = float, help ='specifies the sampling rate with which the continuous signal is discretized (default=10Hz). This directly affects the highest captured frequency (Nyquist). The value is specified in Hz. In case this value is set to -1, the auto mode is launched which sets the sampling frequency automatically to the smallest change in the bandwidth detected. Note that the lowest allowed frequency in the auto mode is 2000 Hz')
        parser.add_argument('-e', '--engine',         type = str, help = 'plot engine. Either plotly (default) or mathplotlib. Specifies the engine used to display the figures. Plotly is used to generate HTML files')
        parser.set_defaults(engine = 'plotly')

    #! FTIO and Predictor Settings
    if 'ftio' in name.lower() or 'predictor' in name.lower():
        parser.set_defaults(mode='sync write')
        parser.add_argument('-f', '--freq', dest='freq', type = float, help ='specifies the sampling rate with which the continuous signal is discretized (default=10Hz). This directly affects the highest captured frequency (Nyquist). The value is specified in Hz. In case this value is set to -1, the auto mode is launched which sets the sampling frequency automatically to the smallest change in the bandwidth detected. Note that the lowest allowed frequency in the auto mode is 2000 Hz')
        parser.set_defaults(freq = 10)
        parser.add_argument('-ts', '--ts',         type = float, help = 'modifies the start time of the examined time window')
        parser.add_argument('-te', '--te',         type = float, help = 'modifies the end time of the examined time window')
        parser.add_argument('-tr', '--transformation', dest='transformation',  type = str, help = 'Specifies the frequency technique to use. Supported modes are: dft (default), wave_disc, and wave_cont')
        parser.set_defaults(transformation='dft')
        parser.add_argument('-e', '--engine',         type = str, help = 'specifies the engine used to display the figures. Either plotly (default) or mathplotlib can be used.  Plotly is used to generate interactive plots as HTML files. Set this value to no if you do not want to generate plots')
        parser.set_defaults(engine = 'plotly')
        parser.add_argument('-o', '--outlier',         type = str, help = 'outlier detection method: Z-score (default), DB-Scan, Isolation_forest, or LOF')
        parser.set_defaults(outlier = 'Z-score')
        parser.add_argument('-le', '--level', dest='level', type = float, help ='specifies the decomposition level for the discrete wavelet transformation (default=3). If specified as auto, the maximum decomposition level is automatic calculated')
        parser.set_defaults(level = 3)
        parser.add_argument('-t', '--tol', dest= 'tol',   type = float, help ='tolerance value')
        parser.set_defaults(tol =  0.8)
        parser.add_argument('-d', '--dtw', action='store_true', help ='performs dynamic time wrapping on the top 3 frequencies (highest contribution) calculated using the DFT if set (default=False)')
        parser.set_defaults(dtw=False)
        parser.add_argument('-re', '--reconstruction', action='store_true', help ='plots reconstruction of top 10 signals on figure')
        parser.set_defaults(reconstruction=False)
        parser.add_argument('-np','--no-psd', dest='psd', action='store_false', help='if set, replace the power density spectrum (a*a/N) with the amplitude spectrum (a)')
        parser.set_defaults(psd=True)
        parser.add_argument('-n', '--n_freq', dest= 'n_freq',   type = float, help ='number of frequencies to extract. By default FTIO finds the dominant frequency. With this flag, up to "n_freq" can be extracted from FTIO')
        parser.set_defaults(n_freq =  0)
        parser.add_argument('-c', '--autocorrelation', dest='autocorrelation', action='store_true', help ='if set, autocorrelation is calculated in addition to DFT. The results are merged to a single prediction at the end')
        parser.set_defaults(autocorrelation=False)
        parser.add_argument('-w', '--window_adaptation', dest='window_adaptation', action='store_true', help ='online time window adaptation. If set to true, the time window is shifted on X hits to X times the previous phases from the current instance. X corresponds to frequency_hits')
        parser.set_defaults(window_adaptation=False)
        parser.add_argument('-fh', '--frequency_hits', dest= 'frequency_hits',   type = float, help ='specifies the number of hits needed to adapt the time window. A hit occurs once a dominant frequency is found')
        parser.set_defaults(frequency_hits =  3)
        parser.add_argument('-v', '--verbose', dest= 'verbose',   action = 'store_true', help ='sets verbose on or off (default=False)')
        parser.set_defaults(verbose =  False)
        parser.add_argument('--zmq', action='store_true', help='avoids opening the generated HTML file since zmq is used')
        parser.set_defaults(zmq=False)
        parser.add_argument('--zmq_source',         type = str, help = 'the source of zmq: TMIO, direct, etc.')
        parser.set_defaults(zmq_source = 'direct')
        parser.add_argument('--zmq_address', '--zmq_address', dest='zmq_address', type = str, help = 'zmq address for communication')
        parser.set_defaults(zmq_address = '*')
        parser.add_argument('--zmq_port', '--zmq_port', dest='zmq_port', type = str, help = 'zmq port for communication')
        parser.set_defaults(zmq_port = '5555')


    #! IOPLOT Settings
    if 'plot' in name.lower():
        parser.set_defaults(mode='')
        parser.add_argument('-z', '--zoom',       type = float, help ='upper zoom limit on the y-axis')
        parser.add_argument('-nt', '--no-threaded', dest='threaded', action='store_false', help= 'turn multithreading off (default=on)')
        parser.set_defaults(threaded=True)
        parser.add_argument('-e', '--engine',         type = str, help = 'plot engine to use. Either plotly (default), dash, or matplotlib')
        parser.set_defaults(engine = 'plotly')
        parser.add_argument('--n_shown_samples', type=int, help='only for dash: Number of shown samples per trace (default: 20_000). Caution: Too small numbers could lead to incorrect representations!')
        parser.set_defaults(n_shown_samples=20_000)
        parser.add_argument('--merge_plots', action='store_true', help='only for dash: Merges the plots to one plot for each io mode. Note: The file dropdown menu then has no functionality')
        parser.set_defaults(merge_plots=False)
        parser.add_argument('--no_disp', action='store_true', help='avoids opening the generated HTML file')
        parser.set_defaults(no_disp=False)

    #! PARSE Settings
    if 'parse' in name.lower():
        parser.add_argument('--scale',  action='store_true', help ='scales the Y-axis')
        parser.set_defaults(scale=False)    

    #! Data modes (for all)
    parser.add_argument('--sum', action='store_true', help ='sum plot: True (default) or False')
    parser.add_argument('--no_sum', dest='sum', action='store_false')
    parser.set_defaults(sum=True)
    parser.add_argument('--avr', action='store_true', help ='avr plot: True (default) or False')
    parser.add_argument('--no_avr', dest='avr', action='store_false')
    parser.set_defaults(avr=True)
    parser.add_argument('--ind', action='store_true', help ='ind plot: True or False (default)')
    parser.add_argument('--no_ind', dest='ind', action='store_false')
    parser.set_defaults(ind=False)
    parser.add_argument('-cf', '--custom_file',   type = str, help = 'passes a [path/filename.py] file containing the translation and pattern for a custom file format similar to:\n https://github.com/tuda-parallel/FTIO/blob/main/examples/custom/txt/custom_input.py')
    parser.set_defaults(custom_file = '')
    parser.add_argument('-x', '--dxt_mode', dest='dxt_mode',  type = str ,  help ='select data to extract from Darshan traces (DXT_POSIX or DXT_MPIIO (default))')
    parser.set_defaults(dxt_mode='DXT_MPIIO')
    parser.add_argument('-l', '--limit',      type = int,   help ='max ranks to consider when reading a folder')
    parser.set_defaults(limit=-1)


    args = parser.parse_args(argv)
    #default values:


    return args



"""Groups predictions according to frequency step
"""

from __future__ import annotations
import numpy as np
from sklearn.cluster import DBSCAN
from ftio.prediction.helper import get_dominant

def group_step(data: list[dict]) -> tuple[list[dict], int]:
    """generates dict contacting predictions. Additionally the entries are grouped according to the frequency resolution between the predictions.

    Args:
        data (dict): predictions

    Returns:
        out (dict): data appended with a group field
        counter (int): Maximal number of groups
    """
    freq_new = 0
    freq_old = 0
    old_window = 0
    time_window = 0
    counter = 0
    out = []

    
    # Method 1: compare frequencies to their next neighbors
    for prediction in data:
        time_window = prediction["t_end"] - prediction["t_start"]
        if len(prediction["dominant_freq"]) >= 1:
            freq_new = prediction["dominant_freq"][0]
            if freq_old != 0 and old_window != 0:
                if abs(freq_old - freq_new) <= 2 * abs(
                    1 / time_window - 1 / old_window
                ):  # Same Frequency!
                    pass
                else:  # Different Frequency!
                    counter += 1
                out.append({**prediction, "group": counter})
            freq_old = freq_new
            old_window = time_window

    return out, counter


def group_dbscan(data: list[dict]) -> tuple[list[dict], int]:
    """generates dict contacting predictions. Additionally the entries are grouped according to DBSCAN with the frequency resolution as the eps distance.

    Args:
        data (dict): predictions

    Returns:
        out (dict): data appended with a group field
        counter (int): Maximal number of groups
    """
    freq = []
    window = []
    tol_max = 0
    tol_min = np.inf
    old_window = 0
    time_window = 0
    counter = 0
    out = []

    for prediction in data:
        time_window = prediction["t_end"] - prediction["t_start"]
        if len(prediction["dominant_freq"]) >= 1:
            res = 1 / time_window - 1 / old_window if old_window != 0 else 0
            tol_max = max(abs(res), tol_max)
            # tol_min = min(abs(res), tol_min)
            freq.append(get_dominant(prediction))
            window.append(time_window)
            out.append(prediction)
            old_window = time_window
    
    tol_min = 1/np.std(window) if window and np.std(window) != 0 else 1e-8
    tol = 2*tol_max if tol_max < 3*tol_min else np.abs(1-(tol_min/np.mean(window)))*tol_max#3 times std means 99 points
    tol = tol if tol > 0 and tol != np.inf else 1e-8 #dbscan expects tol > 0
    # print(f"tol_min is: {tol_min}\ntol_max is: {tol_max}\ntol is: {tol}")

    if out:
        if len(out) == 1:
            counter = 0
            out[0]["group"] = counter
        else:
            X = np.column_stack((freq, freq))
            model = DBSCAN(eps=tol, min_samples=2).fit(X)
            counter = max(model.labels_)
            # print(f"X is {X}\n model.labels_ is {model.labels_}\nout is {out}")
            # for i,f in enumerate(freq):
            #     print(f"freq {f}\nmodel.labels_ is {model.labels_[i]}\ntol is {tol}")
            for i in range(0, len(out)):
                out[i]["group"] = model.labels_[i]

    return out, counter




import matplotlib.pyplot as plt
import plotly.graph_objects as go


class StackPlot:
    def __init__(self, args):
        self.engine = "plotly"
        if args.engine:
            self.engine = args.engine
            if "plotly" in args.engine:
                self.fig = go.Figure()

            if "mat" in args.engine:
                self.b = []
                self.t = []

    def add(self, b, t, name="", group=0):
        if "plotly" in self.engine:
            txt = {}
            if group != 0:
                txt = {"legendgroup": group, "legendgrouptitle_text": group}
            self.fig.add_trace(
                go.Scatter(x=t, y=b, name=name, **txt, line={"shape": "hv", "width": 0.5}, fill="tozeroy", hoverinfo="x+y", stackgroup="one")
            )
        elif "mat" in self.engine:
            if group != 0:
                label = "ranks %i -> %s" % (group, name)
            else:
                label = name
            plt.plot([], [], label=label)
            if not self.b:
                self.t = t
            self.b.append(b)

    def finilize(self):
        if "plotly" in self.engine:
            # template = "plotly_dark"
            template = "plotly"
            width = 1100
            height = 600
            self.fig.update_layout(
                xaxis_title="Time (s)",
                yaxis_title="Bandwidth (B/s)",
                font=dict(family="Courier New, monospace", size=24),
                width=width,
                height=height / 1.1,
                template=template,
            )
            self.fig.update_layout(legend=dict(groupclick="toggleitem"))
            self.fig.show()
        elif "mat" in self.engine:
            plt.stackplot(self.t, self.b, baseline="zero")
            plt.legend()
            plt.xlabel("Time (s)")
            plt.ylabel("Bandwidth (B/s)")
            plt.show()



"""This functions plots error bars (normed and not normed)
"""
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from ftio.plot.helper import format_plot


def make_sub():
    f = make_subplots(
        rows=2,
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.05,
        row_heights=[0.65, 0.35],
    )
    f.update_xaxes(minor=dict(ticklen=6, tickcolor="black", showgrid=True))

    return f


def plot_error_bar(df, s, f=[]):
    if not f:
        f = make_sub()
    modes = ["max", "median"]
    # modes = ['max', 'min', 'median', 'hmean', 'amean', 'whmean']
    x_unqiue = pd.unique(df["number_of_ranks"])
    colors = px.colors.qualitative.Plotly
    if "T" in s:
        symbols = ["circle", "square", "cross", "star-triangle-down", "hourglass"]
        dash = "dash"
    else:
        symbols = ["hexagon", "diamond", "x", "star-triangle-up", "bowtie"]
        # colors = colors[5:]
        colors = colors[2:]
        dash = None

    markeredgecolor = "DarkSlateGrey"
    y = np.zeros((len(modes), len(x_unqiue)))
    y_plus = np.zeros((len(modes), len(x_unqiue)))
    y_minus = np.zeros((len(modes), len(x_unqiue)))
    y_plus_normed = np.zeros((len(modes), len(x_unqiue)))
    y_minus_normed = np.zeros((len(modes), len(x_unqiue)))

    for j in modes:
        for k in range(0, len(x_unqiue)):
            data = df[df["number_of_ranks"] == x_unqiue[k]][j]
            data = data[data != 0]
            if not data.empty:
                y[modes.index(j), k] = data.mean()
                tmp = data - y[modes.index(j), k]
                y_plus[modes.index(j), k] = abs(max(tmp))
                y_minus[modes.index(j), k] = abs(min(tmp))
            if y[modes.index(j), k] != 0:
                y_plus_normed[modes.index(j), k] = (
                    abs(max(tmp)) / y[modes.index(j), k] * 100
                )
                y_minus_normed[modes.index(j), k] = (
                    abs(min(tmp)) / y[modes.index(j), k] * 100
                )
            else:
                y_plus_normed[modes.index(j), k] = 0
                y_minus_normed[modes.index(j), k] = 0
        f.add_trace(
            go.Scatter(
                x=x_unqiue,
                y=y[modes.index(j)],
                error_y=dict(
                    type="data",
                    symmetric=False,
                    array=y_plus[modes.index(j)],
                    arrayminus=y_minus[modes.index(j)],
                ),
                mode="lines+markers",
                name="$\\text{%s}(%s)$" % (j, s),
                legendgroup="%s_%s" % (j, s),
                marker=dict(
                    symbol=symbols[modes.index(j)],
                    line=dict(width=1, color=markeredgecolor),
                ),
                marker_color=colors[modes.index(j)],
                line=dict(dash=dash),
            ),
            row=1,
            col=1,
        )

    f.update_yaxes(title_text="Transfer rate (B/s)", row=1, col=1)
    f.update_xaxes(type="log", row=1, col=1, showspikes=True)
    # f.update_xaxes(type="log",row=1, col=1, showspikes=True) #, tickmode = 'array', tickvals = x_unqiue, ticktext = x_unqiue.astype(str))
    f.update_layout(
        hovermode="x",
        legend_tracegroupgap=1,
        width=900,
        height=400,
        title="Bandwidth with error bars",
    )

    for j in modes:
        f.add_trace(
            go.Scatter(
                x=x_unqiue,
                y=np.repeat(0, len(y[modes.index(j)])),
                error_y=dict(
                    type="data",
                    symmetric=False,
                    array=y_plus_normed[modes.index(j)],
                    arrayminus=y_minus_normed[modes.index(j)],
                ),
                mode="lines+markers",
                name="$\\text{%s}(%s)$" % (j, s),
                showlegend=False,
                legendgroup="%s_%s" % (j, s),
                marker=dict(
                    symbol=symbols[modes.index(j)],
                    line=dict(width=1, color=markeredgecolor),
                ),
                marker_color=colors[modes.index(j)],
                line=dict(dash=dash),
            ),
            row=2,
            col=1,
        )

    f.update_xaxes(
        type="log", title_text="Ranks", row=2, col=1, showspikes=True
    )  # ,tickmode = 'array', tickvals = x_unqiue, ticktext = x_unqiue.astype(str))
    f.update_yaxes(title_text="Rel.<br>dev. (%)", row=2, col=1)
    f.update_yaxes(
        minor=dict(ticklen=4, tickcolor="black", showgrid=True, ticks="inside")
    )
    f.update_layout(
        hovermode="x",
        legend_tracegroupgap=1,
        width=900,
        height=400,
        title="Bandwidth with error bars (normed)",
    )

    if "A" in s:
        title = "Error Bar: Average"
    elif "S" in s:
        title = "Error Bar: Sum"
    else:
        title = "Error Bar: Ind"

    f.update_layout(
        yaxis_title="Transfer Rate (B/s)",
        title=title,
        width=900,
        height=560,
    )
    f = format_plot(f)
    # if 'B' in s and 'E' in s:
    # 	f[-1].show()
    # 	f[-1].write_image("%s.pdf"%s)
    return f


# **********************************************************************
# *                       2. plot_time_error_bar
# **********************************************************************
def plot_time_error_bar(df_time, modes, names, colors, symbols, markeredgecolor):
    f = make_subplots(
        rows=2,
        cols=1,
        shared_xaxes=False,
        vertical_spacing=0.2,
        row_heights=[0.68, 0.32],
    )
    x_unqiue = pd.unique(df_time["number_of_ranks"])
    y = np.zeros((len(modes), len(x_unqiue)))
    y_plus = np.zeros((len(modes), len(x_unqiue)))
    y_minus = np.zeros((len(modes), len(x_unqiue)))
    y_plus_normed = np.zeros((len(modes), len(x_unqiue)))
    y_minus_normed = np.zeros((len(modes), len(x_unqiue)))
    for j in modes:
        for count, _ in enumerate(x_unqiue):
            if "+" in j:
                values = j.split("+")
                values = [x.strip() for x in values]
                for q in values:
                    if values.index(q) == 0:
                        data = df_time[df_time["number_of_ranks"] == x_unqiue[count]][q]
                    else:
                        data = (
                            data
                            + df_time[df_time["number_of_ranks"] == x_unqiue[count]][q]
                        )
            elif "/" in j:
                values = j.split("/")
                values = [x.strip() for x in values]
                for q in values:
                    if values.index(q) == 0:
                        data = df_time[df_time["number_of_ranks"] == x_unqiue[count]][q]
                    else:
                        data = (
                            data
                            / df_time[df_time["number_of_ranks"] == x_unqiue[count]][q]
                        )
            else:
                data = df_time[df_time["number_of_ranks"] == x_unqiue[count]][j]
            data = data[data != 0]
            if not data.empty:
                y[modes.index(j), count] = data.mean()
                tmp = data - y[modes.index(j), count]
                y_plus[modes.index(j), count] = abs(max(tmp))
                y_minus[modes.index(j), count] = abs(min(tmp))
                if y[modes.index(j), count] != 0:
                    y_plus_normed[modes.index(j), count] = (
                        abs(max(tmp)) / y[modes.index(j), count] * 100
                    )
                    y_minus_normed[modes.index(j), count] = (
                        abs(min(tmp)) / y[modes.index(j), count] * 100
                    )
                else:
                    y_plus_normed[modes.index(j), count] = 0
                    y_minus_normed[modes.index(j), count] = 0
        f.add_trace(
            go.Scatter(
                x=x_unqiue,
                y=y[modes.index(j)],
                error_y=dict(
                    type="data",
                    symmetric=False,
                    array=y_plus[modes.index(j)],
                    arrayminus=y_minus[modes.index(j)],
                ),
                mode="lines+markers",
                name=names[modes.index(j)],
                legendgroup=names[modes.index(j)],
                marker=dict(
                    symbol=symbols[modes.index(j)],
                    line=dict(width=1, color=markeredgecolor),
                    size=8 if "Total" not in names[modes.index(j)] else 9,
                ),
                marker_color=colors[modes.index(j)],
            ),
            row=1,
            col=1,
        )
    f.update_yaxes(title_text="Time (s)", row=1, col=1, showspikes=True)
    f.update_xaxes(
        type="log", title_text="Ranks", row=1, col=1, showspikes=True
    )  # ,tickmode = 'array', tickvals = x_unqiue, ticktext = x_unqiue.astype(str))
    f.update_layout(
        hovermode="x",
        legend_tracegroupgap=1,
        width=900,
        height=640,
        title="I/O time with error bars",
    )

    #! Error bar normed
    for j in modes:
        f.add_trace(
            go.Scatter(
                x=x_unqiue,
                y=np.repeat(0, len(y[modes.index(j)])),
                error_y=dict(
                    type="data",
                    symmetric=False,
                    array=y_plus_normed[modes.index(j)],
                    arrayminus=y_minus_normed[modes.index(j)],
                ),
                mode="lines+markers",
                name=names[modes.index(j)],
                showlegend=False,
                legendgroup=names[modes.index(j)],
                marker=dict(
                    symbol=symbols[modes.index(j)],
                    line=dict(width=1, color=markeredgecolor),
                    size=8,
                ),
                marker_color=colors[modes.index(j)],
            ),
            row=2,
            col=1,
        )
    f.update_xaxes(type="log", title_text="Ranks", row=2, col=1, showspikes=True)
    f.update_yaxes(title_text="Rel. dev. (%)", row=2, col=1, showspikes=True)
    f.update_layout(
        hovermode="x",
        legend_tracegroupgap=1,
        width=990,
        height=640,
        title="I/O time with error bars (normed)",
    )
    f = format_plot(f)
    return f


from scipy import stats
import pandas as pd
import numpy as np


class Metrics:
    def __init__(self,args):
        self.t_avr  = []
        self.t_sum  = []
        self.t_ind  = []
        self.b_avr  = []
        self.b_sum  = []
        self.b_ind  = []
        self.args   = args


    def add(self,ranks,run, T1, T2, B1=[], B2=[]):
        if self.args.avr:
            self.assign_metric('t_avr', T1['b_overlap_avr'],ranks,run, T1['t_overlap'])
        if self.args.sum:
            self.assign_metric('t_sum', T1['b_overlap_sum'],ranks,run, T1['t_overlap'])
        if self.args.ind:
            self.assign_metric('t_ind', T2['b_overlap_ind'],ranks,run,T2['t_overlap_ind'])
        if not isinstance(B1, list):
            if self.args.avr:
                self.assign_metric('b_avr', B1['b_overlap_avr'],ranks,run,B1['t_overlap'])
            if self.args.sum:
                self.assign_metric('b_sum', B1['b_overlap_sum'],ranks,run,B1['t_overlap'])
            if self.args.ind:
                self.assign_metric('b_ind', B2['b_overlap_ind'],ranks,run,B2['t_overlap_ind'])


    def get_data(self):
        name = ['number_of_ranks', 'run', 'max', 'min', 'median', 'hmean', 'amean', 'whmean']
        self.t_avr = pd.DataFrame(self.t_avr,columns=name)
        self.t_sum = pd.DataFrame(self.t_sum,columns=name)
        self.t_ind = pd.DataFrame(self.t_ind,columns=name)
        self.b_avr = pd.DataFrame(self.b_avr,columns=name)
        self.b_sum = pd.DataFrame(self.b_sum,columns=name)
        self.b_ind = pd.DataFrame(self.b_ind,columns=name)


    def assign_metric(self,name,b,ranks,run,weights=None):
        tmp = add_metric(b,ranks,run,weights)
        if tmp:
            if name ==  't_avr':
                self.t_avr.append(tmp)
            elif name =='t_sum':
                self.t_sum.append(tmp)
            elif name =='t_ind':
                self.t_ind.append(tmp)
            elif name =='b_avr':
                self.b_avr.append(tmp)
            elif name =='b_sum':
                self.b_sum.append(tmp)
            elif name =='b_ind':
                self.b_ind.append(tmp)
            else:
                pass


    def get(self, name, metric):
        if 't_a' in  name:
            return self.t_avr[metric]
        elif 't_s' in  name:
            return self.t_sum[metric]
        elif 't_i' in  name:
            return self.t_ind[metric]
        elif 'b_a' in  name:
            return self.b_avr[metric]
        elif 'b_s' in  name:
            return self.b_sum[metric]
        elif 'b_i' in  name:
            return self.b_ind[metric]
        else:
            pass


def add_metric(b,ranks,run,weights=None): # b is a pandas dataframe
    b_nonzero = b[b>0]
    if not b_nonzero.empty:
        ranks  = ranks
        run    = run
        max    = b_nonzero.max()
        min    = b_nonzero.min()          
        median = b_nonzero.median()
        amean  = b_nonzero.mean()
        hmean  = stats.hmean(b_nonzero,axis=0)
        b = b.reset_index(drop=True)
        b[b<0] = 0
        weights=(np.asarray([*weights,0]) - np.asarray([0, *weights]))[1:]
        weights[-1] = 0
        weights = weights[b>0] 
        b = b[b>0] 
        whmean  = stats.hmean(b,axis=0,weights=weights/sum(weights))
        return [ranks, run, max, min, median, hmean, amean, whmean]
    else:
        return []



class Time: 

    def  __init__(self, data, rank, args): 
        # total
        self.delta_t_agg     = self.assign(data,"delta_t_agg")
        self.delta_t_agg_io  = self.assign(data,"delta_t_agg_io")
        self.delta_t_com     = self.delta_t_agg - self.delta_t_agg_io

        # read
        self.delta_t_sr       = self.assign(data,"delta_t_sr")
        self.delta_t_ara      = self.assign(data,"delta_t_ara")
        self.delta_t_arr      = self.assign(data,"delta_t_arr")
        self.delta_t_ar_lost  = self.assign(data,"delta_t_ar_lost")

        # write
        self.delta_t_sw       = self.assign(data,"delta_t_sw")
        self.delta_t_awa      = self.assign(data,"delta_t_awa")
        self.delta_t_awr      = self.assign(data,"delta_t_awr")
        self.delta_t_aw_lost  = self.assign(data,"delta_t_aw_lost")
        self.file_index       = args.file_index

        # lib overhead
        self.delta_t_overhead              = self.assign(data,"delta_t_overhead")
        if "delta_t_overhead_post_runtime" in data:
            self.delta_t_overhead_post_runtime = self.assign(data,"delta_t_overhead_post_runtime")
        if "delta_t_overhead_peri_runtime" in data:
            self.delta_t_overhead_peri_runtime      = self.assign(data,"delta_t_overhead_peri_runtime")

        self.delta_t_total    = self.delta_t_overhead + self.delta_t_agg
        self.rank = rank

        self.delta_t_rank0                       = self.assign(data,"delta_t_rank0") 
        self.delta_t_rank0_app                   = self.assign(data,"delta_t_rank0_app") 
        self.delta_t_rank0_overhead_post_runtime = self.assign(data,"delta_t_rank0_overhead_post_runtime") 
        self.delta_t_rank0_overhead_peri_runtime = self.assign(data,name="delta_t_rank0_overhead_peri_runtime") 
        # TODO: compatibility mode with earlier versions
        if "delta_t_rank0_overhead_peri_runtime" not in data:
            self.delta_t_rank0_overhead_peri_runtime = self.assign(data,name="delta_t_rank0_overhead_runtime") 

    def get_data(self):
		#! append list is much faster than append data frame
        name0 = []
        data0 = []
        name0.append("number_of_ranks")
        data0.append(self.rank)
        for attr, value in self.__dict__.items():
            name0.append(attr)
            data0.append(value)

        return name0,data0

    def assign(self, data, name):
        if name in data:
            return data[name]
        else:
            return float('NaN')



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

def legend_fix(data, c) -> bool:
    if c == data.nRun - 1:
        return True
    else:
        return False


def format_plot(fig) -> go.Figure:
    """Formats plots

    Args:
        fig (plotly figure): figure from plotly
    """
    fig.update_layout(
        plot_bgcolor="white",
        legend=dict(
            bgcolor="rgba(255,255,255,.99)",
            bordercolor="Black",
            borderwidth=1,
        ),
        font=dict(family="Courier New, monospace", size=17, color="black"),
        margin=dict(r=5),
    )

    fig.update_xaxes(
        ticks="outside",
        ticklen=10,
        showgrid=True,
        mirror=True,
        showline=True,
        linecolor="black",
        gridcolor="lightgrey",
        minor_ticks="outside",
        minor=dict(
            ticklen=2, tickcolor="black", tickmode="auto", nticks=5, showgrid=True
        ),
    )

    fig.update_yaxes(
        ticks="outside",
        ticklen=10,
        showgrid=True,
        mirror=True,
        showline=True,
        linecolor="black",
        gridcolor="lightgrey",
        minor_ticks="outside",
        minor=dict(
            ticklen=2, tickcolor="black", tickmode="auto", nticks=5, showgrid=True
        ),
    )

    return fig


def save_fig(fig, f, path, name):
    length = len(f)
    index = f.index(fig)
    try:
        print(f"   -> Working on {index}/{length-1}")
        fig.write_image(f"{path}/{name}_{index}.pdf")
    except Exception as error:
        print(f"An exception occurred: {error}")
        fig.write_image(f"{index}.pdf")
    print(f"   -> Finished {index}/{length-1}")


def add_fig_row(nRun, onefig=False, specs=None, subplot_titles=None) -> go.Figure:
    if nRun == 1 or onefig == True:
        # self.f_t.append(go.Figure())
        f = make_subplots(rows=1, cols=1, specs=specs)
    else:
        f = make_subplots(rows=nRun, cols=1, specs=specs, subplot_titles=subplot_titles)

    return f


def add_fig_col(
    nRun, onefig=False, specs=None, subplot_titles=None, horizontal_spacing=0.01
) -> go.Figure:
    if nRun == 1 or onefig == True:
        # self.f_t.append(go.Figure())
        f = make_subplots(rows=1, cols=1, specs=specs)
    else:
        subplot_titles = np.arange(0, nRun).astype(str).tolist()
        subplot_titles = ["Run " + x for x in subplot_titles]
        f = make_subplots(
            rows=1,
            cols=nRun,
            specs=specs,
            shared_yaxes=True,
            subplot_titles=subplot_titles,
            horizontal_spacing=horizontal_spacing,
        )

    return f



import numpy as np
from rich.console import Console
import ftio.prediction.group as gp
from ftio.prediction.helper import get_dominant
from ftio.prediction.probability import Probability


def find_probability(data: list[dict], method:str = "db") -> list:
    """Calculates the conditional probability that expresses
    how probable the frequency (event A) is given that the signal
    is periodic occurred (probability B).
    According to Bayes' Theorem, P(A|B) = P(B|A)*P(A)/P(B)
    P(B|A): Probability that the signal is periodic given that it has a frequency A --> 1
    P(A): Probability that the signal has the frequency A
    P(B): Probability that the signal has is periodic

    Args:
        data (dict): contacting predictions

    Returns:
        None
    """
    p_b = 0
    p_a = []
    p_a_given_b = 0
    p_b_given_a = 1
    grouped_prediction = []
    counter = 0
    out = []

    if data:
        if "step" in method:
            grouped_prediction, counter = gp.group_step(data)
        elif "db" in method:
            grouped_prediction, counter = gp.group_dbscan(data)

        for prediction in data:
            if len(prediction["dominant_freq"]) >= 1:
                p_b += 1

        p_b = p_b / len(data) 
        CONSOLE = Console()
        CONSOLE.print(f"[purple][PREDICTOR]:[/] P(periodic) = {p_b*100:.3f}%")

        if len(grouped_prediction) > 0:
            for group in range(0, counter + 1):
                p_a = 0
                f_min = np.inf
                f_max = 0
                for pred in grouped_prediction:
                    # print(pred)
                    # print(f"index is {group}, group is {pred['group']}")
                    if group == pred["group"]:
                        f_min = min(get_dominant(pred), f_min)
                        f_max = max(get_dominant(pred), f_max)
                        # print(f"group: {group}, pred_group: {pred['group']}, freq: {get_dominant(pred):.3f}, f_min: {f_min:.3f}, f_max:{f_max:.3f}")
                        p_a += 1

                p_a = p_a / len(data) if len(data) > 0 else 0
                p_a_given_b = p_b_given_a * p_a / p_b if p_b > 0 else 0

                prob = Probability(f_min, f_max)
                prob.set(p_b, p_a, p_a_given_b, p_b_given_a)
                prob.display()
                out.append(prob)

    return out